"""This file contains functions implementing
Cluster dynamical mean-field theory (CDMFT).
"""

import timeit

import matplotlib.pyplot as plt
import nlopt
import numpy as np
import scipy.optimize

import pyqcm
from pyqcm import qcm


class convergence_manager:
    """
    class managing a convergence test. One or more instances of this class must be used by the CDMFT
    procedure. One instance corresponds to a quantity that is evaluated at each iteration and then
    compared with the corresponding quantities at a number (`depth`) of previous iterations.
    A difference function (`diff_func`) is applied to the current value of the quantity and each of
    its predecessors, and convergence is declared when the result of this difference (a float) is smaller than
    the tolerance (`tol`) for each of the `depth` previous iterations.
    The possible values of the 'name' argument are:
    * 'parameters' : the set of bath parameters is the quantity used.
    * 'GS energy' : the ground state energy of the impurity problem is used. In the case of more than one cluster, the sum is used.
    * 'hybridization' : the hybridization function (or set thereof, in the case of many clusters). The test is based on the norm of the matrix differences, summed over grid frequencies.
    * 'self-energy' : same as above, but using the impurity self-energy instead.
    * 'distance' : the distance function (difference between successive iterations)
    * any one-body or anomalous lattice operator name. The lattice average is used as test value.
    """

    def __init__(self, name, diff_func, tol, depth=2, stdev=False):
        """
        :param str name: name of the convergence test
        :param function diff_func: difference function applied to the objects tested
        :param float tol: tolerance for convergence
        :param int depth: number of previous iterations the comparison is made with
        :param bool stdev: indicates that the convergence is tested using the standard deviation
        """

        self.name = name
        self.diff_func = diff_func
        self.tol = tol
        self.depth = depth
        self.diff = np.zeros(depth)
        self.iter = 0
        self.X = []  # list of test objects
        self.op = None
        self.converged = False
        self.stdev = stdev
        if stdev:
            self.val = np.zeros(32)
            self.std = None
            self.ave = None

    def test(self, x):
        """
        Tests for convergence and stores the test object x in an array for comparisons with 'depth' future iterations

        :param object x: the object containing the current test quantity
        :returns:True if converged, False otherwise
        """
        self.iter += 1
        if self.stdev:
            return self.stdev_test(x)
        converged = True
        if self.iter <= self.depth:
            self.X.append(x)
            return False
        else:
            for i in range(self.depth):
                self.diff[i] = self.diff_func(x, self.X[i])
            for i in range(self.depth):
                if self.diff[i] > self.tol:
                    converged = False
                    break
            for i in range(self.depth - 1):
                self.X[i] = self.X[i + 1]
            self.X[-1] = x
            if converged:
                self.converged = True
                return True
            else:
                self.converged = False
                return False

    def stdev_test(self, x):
        """
        Tests for convergence using the standard deviation of the accumulated sequence of values.
        Appends x to the internal sequence and computes the standard deviation of the mean;
        convergence is declared when this standard deviation falls below self.tol.

        :param float x: the new value to add to the convergence sequence
        :returns: True if the standard deviation of the mean is below the tolerance, False otherwise
        :rtype: bool
        """
        print(self.name + " = {:1.5g} (added to sequence)".format(x))
        if self.iter > len(self.val):
            self.val.resize(len(self.val) + 32)
        self.val[self.iter - 1] = x
        self.std = np.std(self.val[0 : self.iter]) / np.sqrt(self.iter)
        self.ave = np.mean(self.val[0 : self.iter])
        if self.std < self.tol:
            return True
        else:
            return False

    def print(self):
        """
        Prints the status of the convergence (differences are printed, with a number "depth" of previous iterations)
        """
        if self.iter <= self.depth:
            return
        if self.stdev:
            S = "---> ave({:s}) = {:1.7g} +/- {:1.4g}".format(
                self.name, self.ave, self.std
            )
        else:
            S = "differences in " + self.name + " : "
            for i in range(self.depth):
                S += "{:g}\t".format(self.diff[i])
        if self.converged:
            S += " * converged * "
        print(S, flush=True)


class CDMFT:
    """
    class containing the elements of a CDMFT computation. The constructor executes the computation.

    :param [str] varia: list of variational parameters.
    :param float beta: inverse fictitious temperature (for the frequency grid)
    :param float wc: cutoff frequency (for the frequency grid) or, if grid_type = 'legendre', (w1, w2, n), where w1 is a low-frequency boundary, w2 a high-frequency boundary and n the number of points in each of the 3 segments 
    :param str grid_type: type of frequency grid along the imaginary axis : 'sharp', 'ifreq', 'adapt', 'legendre'
    :param bool selfnorm: multiplies the weights of the frequency grid by the norm of the self-energy
    :param int maxiter: maximum number of CDMFT iterations
    :param int miniter: minimum number of CDMFT iterations
    :param float accur_bath: the x-tolerance for distance function optimization
    :param [str] convergence: the convergence tests (sequence of strings or single string)
    :param [float] accur: the tolerance for the various convergence tests (sequence of floats or single float)
    :param [float] accur_dist: relative tolerance of the distance function when minimizing it
    :param [hartree] hartree: mean-field hartree couplings to incorportate in the convergence procedure
    :param bool converge_with_stdev: If True, checks convergence using the standard deviation of the convergence tests, not the difference
    :param str iteration: method of iteration of parameters ('fixed_point' or 'broyden')
    :param float alpha: if iteration='fixed_point', damping parameter (fraction of the previous iteration in the new one). If iteration='broyden', 1+alpha is the inverse initial Jacobian (or alpha can literally be a matrix, the inverse Jacobian from a previous run).
    :param str method: method to use, as used in scipy.optimize.minimize(). Choices: 'Nelder-Mead' (default), 'Powell', 'CG', 'BFGS', 'ANNEAL', or a choice of NLopt methods : 'NELDERMEAD', 'COBYLA', 'BOBYQA', 'PRAXIS', 'SUBPLEX'
    :param str file: name of the file where the solution is written
    :param str iter_file: name of the file where the CDMFT iterations are recorded
    :param int eps_algo: number of elements in the epsilon algorithm convergence accelerator = 2*eps_algo + 1 (0 = no acceleration)
    :param float initial_step: initial step in the minimization routine
    :param bool SEF: if True, computes the Potthoff functional at the end
    :param bool check_ground_state: if True, checks the ground state consistency and raises exception if inconsistent
    :param int max_function_eval: maximum number of distance function evaluations when minimizing distance
    :param bool compute_potential_energy: If True, computes Tr(Sigma*G) along with the averages
    :param ndarray host_function: if not None, function that computes the host array and passes it to qcm
    :param function pre_host: function to be executed before computing the host. Takes a model instance as argument
    :param float max_value: maximum absolute value of variational parameters
    :ivar lattice_model model: (unique) model on which the computation is based
    :ivar ndarray Hyb: host function
    :ivar ndarray Hyb_down: host function for the spin down component in the case of mixing=4
    :ivar I: current model instance (changes in the course of the computation)

    """

    def __init__(
        self,
        model,
        varia,
        beta=50,
        wc=2.0,
        grid_type="sharp",
        selfnorm=False,
        maxiter=32,
        miniter=0,
        convergence="parameters",
        depth=2,
        accur_bath=1e-6,
        accur=1e-4,
        accur_dist=1e-8,
        hartree=None,
        converge_with_stdev=False,
        iteration="broyden",  # or 'fixed_point'
        alpha=0.0,
        method="Nelder-Mead",
        file="cdmft.tsv",
        iter_file="cdmft_iter.tsv",
        eps_algo=0,
        initial_step=0.1,
        SEF=False,
        check_ground_state=False,
        max_function_eval=1_000_000,
        compute_potential_energy=False,
        host_function=None,
        pre_host=None,
        max_value=100,
    ):

        self.accur_bath = accur_bath
        self.accur_dist = accur_dist
        self.alpha = alpha
        self.beta = beta
        self.check_ground_state = check_ground_state
        self.dist = 1e6
        self.hartree = hartree
        self.grid_type = grid_type
        self.selfnorm = selfnorm
        self.host_function = host_function
        self.Hyb = None  # internal : hybridization function
        self.Hyb_down = (
            None  # internal : hybridization function (spin downs, when mixing=4)
        )
        self.initial_step = initial_step
        self.iter_file = iter_file
        self.max_function_eval = max_function_eval
        self.max_value = max_value
        self.method = method
        self.model = model
        self.pre_host = pre_host
        self.sigma = None  # internal : self-energy
        self.sigma_down = None  # internal : self-energy (spin downs, when mixing=4)
        self.varia = varia
        self.wc = wc

        if pyqcm.is_sequence(accur) == False:
            accur = (accur,)

        if pyqcm.is_sequence(varia) == False:
            raise ValueError(
                "The argument 'varia' of CDMFT should be a list of variational parameters"
            )

        if iteration not in ["broyden", "fixed_point"]:
            raise ValueError(
                "the argument 'iteration' of CDMFT() should be one of 'broyden', 'fixed_point'"
            )

        # ----------- Managing the variational parameters ---------

        if pyqcm.is_sequence(hartree) == False and hartree is not None:
            self.hartree = (hartree,)
        else:
            self.hartree = hartree
        if self.hartree is None:
            self.hartree = []

        var = [[] for i in range(model.nclus + 1)]
        for v in varia:
            p = v.partition("_")
            if len(p) == 1:
                raise ValueError(
                    "{:s} cannot be a CDMFT variational parameter".format(v)
                )
            else:
                var[model.sys_clus[int(p[2]) - 1] + 1].append(v)
        var[0] = [h.Vm for h in self.hartree]
        self.nvar = np.zeros(model.nclus + 1, int)
        for c in range(model.nclus + 1):
            self.nvar[c] = len(var[c])

        qcm.CDMFT_variational_set(var[1:])

        self.varia = []
        for v in var:
            self.varia += v
        self.nvaria = len(self.varia)
        self.x = np.empty(self.nvaria)
        self.var_data = np.empty((self.nvaria, maxiter + 1))

        P = model.parameters()
        for i, v in enumerate(self.varia):
            self.x[i] = P[v]

        # ------------------------- convergence test initialization ------------------------
        if pyqcm.is_sequence(convergence) == False:
            convergence = (convergence,)
        if len(convergence) != len(accur):
            raise ValueError(
                'The number of convergence tests must be the same as the length of "accur"'
            )
        self.convergence_test = []
        convergence_test_string = ""
        for i, C in enumerate(convergence):
            convergence_test_string += C + "/"
            if C in model.parameters():
                conv_manager = convergence_manager(
                    C,
                    lambda x, y: np.abs(x - y),
                    accur[i],
                    depth,
                    stdev=converge_with_stdev,
                )
                conv_manager.op = C
            else:
                if converge_with_stdev and C != "GS energy":
                    raise ValueError(
                        'In CDMFT, converge_with_stdev=True is only possible when "convergence" is the name of an operator or GS energy'
                    )
                op = None
                if C == "parameters":
                    conv_manager = convergence_manager(
                        C,
                        lambda x, y: np.linalg.norm(x - y) / x.shape[0],
                        accur[i],
                        depth,
                    )
                elif C == "hybridization":
                    conv_manager = convergence_manager(
                        C, lambda x, y: self.diff_matrix(x, y), accur[i], depth
                    )
                elif C == "self-energy":
                    if grid_type == "adapt":
                        raise ValueError(
                            "cannot use convergence on self-energy with adaptive frequency grid"
                        )
                    conv_manager = convergence_manager(
                        C, lambda x, y: self.diff_matrix(x, y), accur[i], depth
                    )
                elif C == "GS energy":
                    conv_manager = convergence_manager(
                        C,
                        lambda x, y: np.abs(x - y),
                        accur[i],
                        depth,
                        stdev=converge_with_stdev,
                    )
                elif C == "distance":
                    conv_manager = convergence_manager(
                        C, lambda x, y: np.abs(x - y) / y, accur[i], depth
                    )
                else:
                    raise ValueError(
                        'type of convergence test "' + C + '" in CDMFT does not exist'
                    )
            self.convergence_test.append(conv_manager)
        convergence_test_string = convergence_test_string[:-1]

        # ----------- Banner (depending on the presence of hartree parameters) -------------
        if hartree:
            pyqcm.banner(
                "CDMFT procedure (combined with Hartree procedure)", "*", skip=1
            )
        else:
            pyqcm.banner("CDMFT procedure", "*", skip=1)

        # -------------- first define the frequency grid for the distance function ---------
        print("frequency grid type = ", grid_type)
        print("fictitious inverse temperature = ", beta)
        print("frequency cutoff = ", wc)
        if alpha is float:
            print("damping factor = ", self.alpha)
        print("-" * 100)

        # ------------------------------- CDMFT main iteration loop ------------------------
        self.niter = 0
        self.iter = 0

        def F(x):
            self.x = np.copy(x)
            self.CDMFT_step()
            return x - self.x

        def G():
            return self.check_convergence()

        if iteration == "broyden":
            try:
                actual_method = "broyden"
                self.x, self.niter, self.alpha = pyqcm.broyden(
                    F,
                    self.x,
                    self.alpha,
                    maxiter=maxiter,
                    miniter=miniter,
                    xtol=1e-6,
                    convergence_test=G,
                )
            except Exception as E:
                raise pyqcm.SolverError("Failure of the CDMFT method (broyden)") from E

        elif iteration == "fixed_point":
            try:
                actual_method = "fixed_point"
                self.x, self.niter = pyqcm.fixed_point_iteration(
                    F,
                    self.x,
                    xtol=1e-6,
                    convergence_test=G,
                    maxiter=maxiter,
                    miniter=miniter,
                    alpha=self.alpha,
                    eps_algo=eps_algo,
                )
            except Exception as E:
                raise pyqcm.SolverError(
                    "Failure of the CDMFT method (fixed-point)"
                ) from E

        else:
            raise ValueError(
                'unknown iteration method in CDMFT. must be either "broyden" or "fixed_point". Check spelling.'
            )

        # Here we have converged
        self.I = pyqcm.model_instance(
            self.model
        )  # a last instance with the converged parameters

        # check consistency
        self.I.GS_consistency(self.check_ground_state)
        var_val = pyqcm.varia_table(self.varia, self.x)
        pyqcm.banner(
            "converged variational parameters ({:d} iterations)".format(self.niter), "-"
        )
        print(var_val)

        ave = self.I.averages(pr=True)
        if compute_potential_energy:
            self.I.potential_energy()
        if SEF:
            omega = self.I.Potthoff_functional(self.hartree)

        if file != None:
            self.I.props["opt_method"] = method
            self.I.props["CDMFT_method"] = actual_method
            self.I.props["CDMFT_iterations"] = self.niter
            self.I.props["dist_function"] = self.grid.dist_function
            self.I.props["convergence"] = convergence_test_string
            self.I.props["min_dist"] = self.dist
            self.I.write_summary(file)

        pyqcm.banner("CDMFT completed successfully", "*")

    # -----------------------------------------------------------------------------------------------
    def CDMFT_step(self):
        """
        Performs a CDMFT step that brings the system from the current values (self.x) of the bath and Hartree
        parameters and updates it to the next set of values
        """

        check_bounds(self.x, self.max_value, v=self.varia)

        self.model.set_parameter(self.varia, self.x)
        self.I = pyqcm.model_instance(self.model)

        if (
            self.grid_type == "adapt"
        ):  # find the maximum value of variational (bath) parameters
            self.wc = max([2.0, max(self.x[self.nvar[0] :])])

        self.grid = frequency_grid(self.I, self.grid_type, self.beta, self.wc, self.selfnorm)

        # solve the impurity problem

        # initializes G_host
        t1 = timeit.default_timer()
        if self.pre_host != None:
            self.pre_host(self.I)

        # computing or transferring the host array --------------------------------------

        if self.host_function == None:
            qcm.CDMFT_host(self.grid.wr, self.grid.weight, self.I.label)
        else:
            self.host_function(self.I)
        # --------------------------------------------------------------------------------

        self.I.GS_consistency(self.check_ground_state)

        self.set_Hyb()
        t2 = timeit.default_timer()
        time_ED = t2 - t1

        # updates the Hartree mean-field parameters
        if self.nvar[0]:
            for C in self.hartree:
                C.update(self.I, pr=True)
            P = self.model.parameters()
            for i, h in enumerate(self.hartree):
                self.x[i] = P[h.Vm]

        gs = self.I.ground_state()

        # --------------------------------------------------------------------------------
        # optimization of the bath parameters, one cluster at a time

        self.dist = 0.0
        ic = self.nvar[0]
        for c in range(1, self.model.nclus + 1):
            if self.nvar[c] == 0:
                continue
            x0 = self.x[ic : ic + self.nvar[c]]
            DISTANCE = lambda x, grad=None: qcm.CDMFT_distance(x, c - 1, self.I.label)
            sol = optimize(
                DISTANCE,
                x0,
                self.method,
                self.initial_step,
                self.accur_bath,
                self.accur_dist,
                self.max_function_eval,
            )
            opt_x, opt_iter_done, opt_success, opt_fun = sol

            self.dist += opt_fun

            if self.method != "ANNEAL" and not opt_success:
                print(sol)
                pyqcm.banner(
                    "Error in the scipy minimization procedure within CDMFT ", "!"
                )
                raise pyqcm.MinimizationError()

            if opt_iter_done > self.max_function_eval:
                print(sol)
                pyqcm.banner(
                    "number of function evaluations exceeds preset maximum of {:d}".format(
                        self.max_function_eval
                    ),
                    "!",
                )
                raise pyqcm.MinimizationError()

            if np.any(np.isnan(opt_x)):
                print(sol)
                pyqcm.banner("NaN found in optimization of bath parameters", "!")
                raise pyqcm.MinimizationError(
                    "NaN found in optimization of bath parameters"
                )

            # push back into array
            self.x[ic : ic + self.nvar[c]] = np.copy(opt_x)
            self.var_data[ic : ic + self.nvar[c], self.iter] = np.copy(opt_x)

            check_bounds(opt_x, self.max_value, v=self.varia[ic : ic + self.nvar[c]])

            ic += self.nvar[c]

        t3 = timeit.default_timer()
        time_MIN = t3 - t2

        # writing the parameters in a progress file
        self.I.props["opt_method"] = self.method
        self.I.props["dist_function"] = self.grid.dist_function
        self.I.props["min_dist"] = self.dist
        self.I.write_summary(self.iter_file)

        print("GS sector : ", [X[1] for X in gs])
        print(
            "{:d} minimization steps, time(MIN)/time(ED)={:.3g}, distance = {:1.4g}".format(
                opt_iter_done, time_MIN / time_ED, opt_fun
            ),
            flush=True,
        )

        var_val = pyqcm.varia_table(self.varia, self.x)
        print("updated variational parameters:\n{:s}".format(var_val))
        self.iter += 1

        return

    # -----------------------------------------------------------------------------------------------
    def check_convergence(self):
        """
        Performs the CDMFT convergence tests
        returns True if all are converged, otherwise False
        """

        converged = True
        for C in self.convergence_test:
            if C.name == "GS energy":
                gs = self.I.ground_state()
                E0 = 0.0
                for x in gs:
                    E0 += x[0]
                converged = converged and C.test(E0)

            elif C.name == "self-energy":
                self.set_sigma()
                if self.model.mixing == 4:
                    T = C.test((self.sigma, self.sigma_down))
                    converged = converged and T
                else:
                    T = C.test(self.sigma)
                    converged = converged and T

            elif C.name == "hybridization":
                if self.model.mixing == 4:
                    T = C.test((self.Hyb, self.Hyb_down))
                    converged = converged and T
                else:
                    T = C.test(self.Hyb)
                    converged = converged and T

            elif C.name == "parameters":
                T = C.test(np.copy(self.x))
                converged = converged and T

            elif C.name == "distance":
                T = C.test(self.dist)
                converged = converged and T

            elif C.op != None:
                T = C.test(self.I.averages()[C.op])
                converged = converged and T

        for C in self.convergence_test:
            C.print()

        if converged:
            pyqcm.banner("CDMFT has converged", "=")
            for C in self.convergence_test:
                C.print()

        return converged

    # -----------------------------------------------------------------------------------------------
    def set_Hyb(self):
        """Computes the hybridization function, i.e.
        an array of arrays of matrices.
        Hyb[i], for cluster #i, is a (nw,d,d) Numpy array. with nw frequencies, and d sites
        Hyb_down[i] is the same, for the spin-down part, if mixing=4

        :returns None

        """
        self.Hyb0 = self.Hyb
        self.Hyb = []
        for j in range(self.model.nclus):
            d = self.model.dimGFC[j]
            self.Hyb.append(np.zeros((self.grid.nw, d, d), dtype=np.complex128))

        for i in range(self.grid.nw):
            for c in range(self.model.nclus):
                self.Hyb[c][i, :, :] = self.I.hybridization_function(
                    self.grid.w[i], c, spin_down=False
                )

        if self.model.mixing == 4:
            self.Hyb0_down = self.Hyb_down
            self.Hyb_down = []
            for c in range(self.model.nclus):
                d = self.model.dimGFC[c]
                self.Hyb_down.append(
                    np.zeros((self.grid.nw, d, d), dtype=np.complex128)
                )

            for i in range(self.grid.nw):
                for c in range(self.model.nclus):
                    self.Hyb_down[c][i, :, :] = self.I.hybridization_function(
                        self.grid.w[i], c, spin_down=True
                    )

    # -----------------------------------------------------------------------------------------------
    def diff_matrix(self, X, Y):
        """
        Computes a difference in hybridization functions between the current iteration (Hyb) and the previous one (Hyb0)
        :param object X : the current test object
        :param object Y : any past test object (same structure as X)
        :returns float: the difference in hybridization arrays

        """

        diff = 0.0
        g = self.grid
        if self.model.mixing != 4:
            for i in range(g.nw):
                for c in range(self.model.nclus):
                    delta = X[c][i, :, :] - Y[c][i, :, :]
                    norm = np.linalg.norm(delta)
                    diff += g.weight[i] * norm * norm

        if self.model.mixing == 4:
            for i in range(g.nw):
                for c in range(self.model.nclus):
                    delta = X[0][c][i, :, :] - Y[0][c][i, :, :]
                    norm = np.linalg.norm(delta)
                    diff += g.weight[i] * norm * norm
            for i in range(g.nw):
                for c in range(self.model.nclus):
                    delta = X[1][c][i, :, :] - Y[1][c][i, :, :]
                    norm = np.linalg.norm(delta)
                    diff += g.weight[i] * norm * norm

        if self.model.mixing == 0:
            diff *= 2
        elif self.model.mixing == 3:
            diff /= 2
        diff /= g.nw
        return np.sqrt(diff)

    # -----------------------------------------------------------------------------------------------
    def set_sigma(self):
        """Computes the self-energy on the frequency grid
        an array of arrays of matrices. sigma[i], for cluster #i, is a (nw,d,d) Numpy array. with nw frequencies, and d sites

        :returns None

        """
        self.sigma0 = self.sigma
        self.sigma = []
        for c in range(self.model.nclus):
            d = self.model.dimGFC[c]
            self.sigma.append(np.zeros((self.grid.nw, d, d), dtype=np.complex128))

        for i in range(self.grid.nw):
            for c in range(self.model.nclus):
                self.sigma[c][i, :, :] = self.I.cluster_self_energy(
                    self.grid.w[i], c, spin_down=False
                )

        if self.model.mixing == 4:
            self.sigma0_down = self.sigma_down
            self.sigma_down = []
            for c in range(self.model.nclus):
                d = self.model.dimGFC[c]
                self.sigma_down.append(
                    np.zeros((self.grid.nw, d, d), dtype=np.complex128)
                )

            for i in range(self.grid.nw):
                for c in range(self.model.nclus):
                    self.sigma_down[c][i, :, :] = self.I.cluster_self_energy(
                        self.grid.w[i], c, spin_down=True
                    )

    # -----------------------------------------------------------------------------------------------
    def plot_iterations(self):
        """
        Plots the variational parameters as a function of iteration, for diagnostics purposes
        """

        ncols = 3
        nrows = 1 + (self.varia.nvar_tot - 1) // ncols
        fig, ax = plt.subplots(nrows, ncols, sharex=True)
        fig.set_size_inches(24 / 2.54, nrows * 6 / 2.54)
        niter = self.var_data.shape[1]
        for i, x in enumerate(self.var):
            if nrows == 1:
                plt.sca(ax[i])
            else:
                plt.sca(ax[i // ncols, i % ncols])
            plt.plot(range(self.niter), self.var_data[i, 0:niter], "o-", ms=3, lw=0.5)
            plt.title(self.varia[i])
        plt.savefig("iterations.pdf")


####################################################################################################
class frequency_grid:
    """
    This class contains the imaginary frequency grid data, including weights

    :param model_instance I: current model instance
    :param str grid_type: type of frequency grid along the imaginary axis : 'sharp', 'ifreq'
    :param float beta: inverse fictitious temperature (for the frequency grid)
    :param float wc: cutoff frequency (for the frequency grid) OR, if grid_type='legendre', (w1, w2, n) where w1 and w2 are low- and high-frequency boundaries and n is the number of points for each of the 3 frequency regions
    :param bool self_norm: if true, multiplies the weight by the norm of the self-energy

    :ivar float beta: inverse fictitious temperature
    :ivar float wc: cutoff frequency (for the frequency grid)
    :ivar str grid_type: type of frequency grid along the imaginary axis : 'sharp', 'ifreq', 'self'
    :ivar wr: array of frequencies (real array, i.e., imaginary part of the frequencies)
    :ivar w: array of complex frequencies (= wr * 1j)
    :ivar [float] weight: array of weights for the different frequencies of the grid
    :ivar int nw: number of frequencies in the grid
    """

    def __init__(self, I=None, grid_type="sharp", beta=50, wc=2, self_norm=False):
        self.beta = beta
        self.wc = wc
        self.grid_type = grid_type

        if grid_type == "legendre":
            assert pyqcm.is_sequence(wc)
            self.wr, self.weight = pyqcm.legendre_frequency_grid(wc[0], wc[1], wc[2])
            self.w = self.wr * 1j
            self.dist_function = "legendre/{:.1f}/{:.1f}/{:d}".format(
                wc[0], wc[1], wc[2]
            )
            self.nw = len(self.wr)
        else:
            self.wr = np.arange(
                (np.pi / self.beta), self.wc + 1e-6, 2 * np.pi / self.beta
            )
            self.nw = len(self.wr)
            self.w = np.ones(len(self.wr), dtype=np.complex128)
            self.w = self.w * 1j
            self.w *= self.wr
            if self.grid_type == "sharp" or self.grid_type == "adapt":
                self.weight = np.ones(self.nw)
                self.weight *= 1.0 / self.nw
                self.dist_function = "{:s}_wc_{:.3g}_b_{:d}".format(
                    self.grid_type, self.wc, int(self.beta)
                )
            elif self.grid_type == "ifreq":
                self.weight = 1.0 / self.wr
                self.weight *= 1.0 / self.weight.sum()
                self.dist_function = "ifreq_wc_{0:.1f}_b_{1:d}".format(
                    self.wc, int(self.beta)
                )
            else:
                raise ValueError(f"unknown frequency grid type `{grid_type}`")
        if self_norm:
            Sig_inf = I.cluster_self_energy(1.0e6j)
            for i, x in enumerate(self.w):
                Sig = I.cluster_self_energy(x) - Sig_inf
                self.weight[i] *= np.linalg.norm(Sig)
            self.weight *= 1.0 / self.weight.sum()
            self.dist_function += "-self"
        



####################################################################################################
class general_bath:
    """
    Class that construct a cluster model with nb bath orbitals, in the most general way possible, with possible restrictions.

    :param str name: name of the cluster-bath model to be defined
    :param int ns: number of sites in the cluster
    :param int nb: number of bath orbitals in the cluster
    :param bool spin_dependent: if True, the parameters are spin dependent
    :param bool spin_flip: if True, spin-flip hybridizations are present
    :param bool singlet: if True, defines anomalous singlet hybridizations
    :param bool triplet: if True, defines anomalous triplet hybridizations
    :param bool complex: if True, defines imaginary parts as well, when appropriate
    :param list[list[int]] sites: 2-level list of sites to couple to the bath orbitals (labels from 1 to ns). Format resembles [[site labels to bind to orbital 1], ...] .

    :ivar int ns: number of physical sites in the cluster
    :ivar int nb: number of bath orbitals in the cluster
    :ivar str name: name of the cluster model
    :ivar [str] var_E: list of bath parameters of the type "energy level"
    :ivar [str] var_H: list of bath parameters of the type "hybridization"
    :ivar bool spin_dependent: True if the bath parameter conserve spin, but are different for spins up and down
    :ivar bool spin_flip: True if we include spin-flip hybridizations terms
    :ivar bool singlet: True if we include anomalous hybridizations of the singlet type
    :ivar bool triplet: True if we include anomalous hybridizations of the triplet type
    :ivar bool complex: True if the model is complex-valued and thus hybridization operators have both real and imaginary components

    """

    def __init__(
        self,
        ns,
        nb,
        name="clus",
        spin_dependent=False,
        spin_flip=False,
        singlet=False,
        triplet=False,
        complex=False,
        sites=None,
    ):
        self.CM = pyqcm.cluster_model(ns, n_bath=nb, name=name)
        self.CM.var_E = []
        self.CM.var_H = []
        self.spin_dependent = spin_dependent
        self.spin_flip = spin_flip
        self.singlet = singlet
        self.triplet = triplet
        self.complex = complex
        no = ns + nb
        if sites is None:
            self.sites = [range(1, ns + 1) for i in range(nb)]
        else:
            if len(sites) != nb:
                print(
                    'the format of the argument "sites" is incorrect : it should be a sequence of ',
                    ns,
                    " sequences",
                )
                raise ValueError(
                    f'the format of the argument "sites" is incorrect : it should be a sequence of {ns} sequences'
                )
            self.sites = sites

        self.nmixed = 1
        if spin_flip:
            self.nmixed *= 2
        if singlet or triplet:
            self.nmixed *= 2

        # bath energies
        if spin_dependent or spin_flip:
            for x in range(1, nb + 1):
                param_name = "eb{:d}u".format(x)
                self.CM.new_operator(param_name, "one-body", [(x + ns, x + ns, 1)])
                self.CM.var_E.append(param_name)
                param_name = "eb{:d}d".format(x)
                self.CM.new_operator(
                    param_name, "one-body", [(x + ns + no, x + ns + no, 1)]
                )
                self.CM.var_E.append(param_name)
        else:
            for x in range(1, nb + 1):
                param_name = "eb{:d}".format(x)
                self.CM.new_operator(
                    param_name,
                    "one-body",
                    [(x + ns, x + ns, 1), (x + ns + no, x + ns + no, 1)],
                )
                self.CM.var_E.append(param_name)

        # hybridizations
        if spin_dependent or spin_flip:
            for x in range(1, nb + 1):
                for y in self.sites[x - 1]:
                    param_name = "tb{:d}u{:d}u".format(x, y)
                    self.CM.new_operator(param_name, "one-body", [(y, x + ns, 1)])
                    self.CM.var_H.append(param_name)
                    param_name = "tb{:d}d{:d}d".format(x, y)
                    self.CM.new_operator(
                        param_name, "one-body", [(y + no, x + ns + no, 1)]
                    )
                    self.CM.var_H.append(param_name)
                    if spin_flip:
                        param_name = "tb{:d}u{:d}d".format(x, y)
                        self.CM.new_operator(
                            param_name, "one-body", [(x + ns, y + no, 1)]
                        )
                        self.CM.var_H.append(param_name)
                        param_name = "tb{:d}d{:d}u".format(x, y)
                        self.CM.new_operator(
                            param_name, "one-body", [(y + no, x + ns + no, 1)]
                        )
                        self.CM.var_H.append(param_name)

                if complex:
                    for y in self.sites[x - 1][1:]:
                        param_name = "tb{:d}u{:d}ui".format(x, y)
                        self.CM.new_operator_complex(
                            param_name, "one-body", [(y, x + ns, 1j)]
                        )
                        self.CM.var_H.append(param_name)
                        param_name = "tb{:d}d{:d}di".format(x, y)
                        self.CM.new_operator_complex(
                            param_name, "one-body", [(y + no, x + ns + no, 1j)]
                        )
                        self.CM.var_H.append(param_name)
                    if spin_flip:
                        for y in self.sites[x - 1]:
                            param_name = "tb{:d}u{:d}di".format(x, y)
                            self.CM.new_operator_complex(
                                param_name, "one-body", [(x + ns, y + no, 1j)]
                            )
                            self.CM.var_H.append(param_name)
                            param_name = "tb{:d}d{:d}ui".format(x, y)
                            self.CM.new_operator_complex(
                                param_name, "one-body", [(y + no, x + ns + no, 1j)]
                            )
                            self.CM.var_H.append(param_name)

        else:
            for x in range(1, nb + 1):
                for y in self.sites[x - 1]:
                    param_name = "tb{:d}{:d}".format(x, y)
                    self.CM.new_operator(
                        param_name,
                        "one-body",
                        [(y, x + ns, 1), (y + no, x + ns + no, 1)],
                    )
                    self.CM.var_H.append(param_name)

                if complex:
                    for y in self.sites[x - 1][1:]:
                        param_name = "tb{:d}{:d}i".format(x, y)
                        self.CM.new_operator_complex(
                            param_name,
                            "one-body",
                            [(y, x + ns, 1j), (y + no, x + ns + no, 1j)],
                        )
                        self.CM.var_H.append(param_name)

        if singlet:
            for x in range(1, nb + 1):
                for y in self.sites[x - 1]:
                    param_name = "sb{:d}{:d}".format(x, y)
                    self.CM.new_operator(
                        param_name,
                        "anomalous",
                        [(y, x + ns + no, 1), (x + ns, y + no, 1)],
                    )
                    self.CM.var_H.append(param_name)
                if complex:
                    for y in self.sites[x - 1]:
                        if x == 1 and y == 1:
                            continue
                        param_name = "sb{:d}{:d}i".format(x, y)
                        self.CM.new_operator_complex(
                            param_name,
                            "anomalous",
                            [(y, x + ns + no, 1j), (x + ns, y + no, 1j)],
                        )
                        self.CM.var_H.append(param_name)

        if triplet:
            for x in range(1, nb + 1):
                for y in self.sites[x - 1]:
                    param_name = "pb{:d}{:d}".format(x, y)
                    self.CM.new_operator(
                        param_name,
                        "anomalous",
                        [(y, x + ns + no, 1), (x + ns, y + no, -1)],
                    )
                    self.CM.var_H.append(param_name)
                if complex:
                    for y in self.sites[x - 1]:
                        if x == 1 and y == 1:
                            continue
                        param_name = "pb{:d}{:d}i".format(x, y)
                        self.CM.new_operator_complex(
                            param_name,
                            "anomalous",
                            [(y, x + ns + no, 1j), (x + ns, y + no, -1j)],
                        )
                        self.CM.var_H.append(param_name)

    # -----------------------------------------------------------------------------------------------
    def __str__(self):
        S = 'cluster model "' + self.name + '"'
        if self.spin_flip:
            S += ", spin flip"
        if self.spin_dependent:
            S += ", spin dependent"
        if self.complex:
            S += ", complex-valued"
        if self.singlet:
            S += ", singlet SC"
        if self.singlet:
            S += ", triplet SC"
        S += "\nbath energies : "
        for x in self.CM.var_E:
            S += x + ", "
        S = S[0:-2] + "\nhybridizations : "
        for x in self.CM.var_H:
            S += x + ", "
        return S[0:-2]

    # -----------------------------------------------------------------------------------------------
    def varia_E(self, c=1):
        """returns a list of parameter names from the bath energies with the suffix appropriate for cluster c

        :param int c: label of the cluster (starts at 1)
        :return [str]: list of parameter names from the bath energies with the suffix appropriate for cluster c

        """
        v = []
        for x in self.CM.var_E:
            v.append(x + "_" + str(c))
        return v

    # -----------------------------------------------------------------------------------------------
    def varia_H(self, c=1):
        """returns a list of parameter names from the bath hybridization with the suffix appropriate for cluster c

        :param int c: label of the cluster (starts at 1)
        :return [str]: list of parameter names from the bath hybridization with the suffix appropriate for cluster c

        """
        v = []
        for x in self.CM.var_H:
            v.append(x + "_" + str(c))
        return v

    # -----------------------------------------------------------------------------------------------
    def varia(self, H=None, E=None, c=1, spin_down=False):
        """creates a dict of variational parameters to values taken from the hybridization matrix H and the energies E, for cluster c

        :param ndarray H: matrix of hybridization values, shape (nmixed*nb, nmixed*ns)
        :param ndarray E: array of energy values for the bath orbitals
        :param int c: label of the cluster (starts at 1), used as suffix in parameter names
        :param bool spin_down: True for the spin-down values
        :return {str,float}: dict of variational parameters to values

        """
        nb = self.nb
        ns = self.ns
        no = ns + nb
        nn = no * self.nmixed // 2

        if H.shape != (self.nmixed * self.nb, self.nmixed * self.ns):
            raise ValueError(
                "shape of hybridization matrix does not match model in general bath back propagation"
            )

        D = {}
        # bath energies
        if self.spin_flip:
            for x in range(1, nb + 1):
                param_name = "eb{:d}u_{:d}".format(x, c)
                D[param_name] = E[x - 1]
                param_name = "eb{:d}d_{:d}".format(x, c)
                D[param_name] = E[x - 1 + nb]

        elif self.spin_flip and not spin_down:
            for x in range(1, nb + 1):
                param_name = "eb{:d}u_{:d}".format(x, c)
                D[param_name] = E[x - 1]
        elif self.spin_flip and spin_down == True:
            for x in range(1, nb + 1):
                param_name = "eb{:d}d_{:d}".format(x, c)
                D[param_name] = E[x - 1]
        else:
            for x in range(1, nb + 1):
                param_name = "eb{:d}_{:d}".format(x, c)
                D[param_name] = E[x - 1]

        # hybridizations
        if self.spin_flip:
            for x in range(1, nb + 1):
                for y in self.sites:
                    param_name = "tb{:d}u{:d}u_{:d}".format(x, y, c)
                    D[param_name] = H[x - 1, y - 1].real
                    param_name = "tb{:d}d{:d}d_{:d}".format(x, y, c)
                    D[param_name] = H[x + nb - 1, y - 1].real
                    param_name = "tb{:d}u{:d}d_{:d}".format(x, y, c)
                    D[param_name] = H[x - 1, y + ns - 1].real
                    param_name = "tb{:d}d{:d}u_{:d}".format(x, y, c)
                    D[param_name] = H[x + nb - 1, y + ns - 1].real

                if self.complex:
                    for y in self.sites[1:]:
                        param_name = "tb{:d}u{:d}ui_{:d}".format(x, y, c)
                        D[param_name] = H[x - 1, y - 1].imag
                        param_name = "tb{:d}d{:d}di_{:d}".format(x, y, c)
                        D[param_name] = H[x + nb - 1, y - 1].imag
                        param_name = "tb{:d}u{:d}di_{:d}".format(x, y, c)
                        D[param_name] = H[x - 1, y + ns - 1].imag
                        param_name = "tb{:d}d{:d}ui_{:d}".format(x, y, c)
                        D[param_name] = H[x + nb - 1, y + ns - 1].imag

        elif self.spin_dependent:
            for x in range(1, nb + 1):
                for y in self.sites:
                    if spin_down:
                        param_name = "tb{:d}d{:d}d_{:d}".format(x, y, c)
                    else:
                        param_name = "tb{:d}u{:d}u_{:d}".format(x, y, c)
                    D[param_name] = H[x - 1, y - 1].real

                if self.complex:
                    for y in self.sites[1:]:
                        if spin_down:
                            param_name = "tb{:d}d{:d}di_{:d}".format(x, y, c)
                        else:
                            param_name = "tb{:d}u{:d}ui_{:d}".format(x, y, c)
                        D[param_name] = H[x - 1, y - 1].imag

        else:
            for x in range(1, nb + 1):
                for y in self.sites:
                    param_name = "tb{:d}{:d}_{:d}".format(x, y, c)
                    D[param_name] = H[x - 1, y - 1].real

                if self.complex:
                    for y in self.sites[1:]:
                        param_name = "tb{:d}{:d}i_{:d}".format(x, y, c)
                        D[param_name] = H[x - 1, y - 1].imag

        if self.singlet:
            for y in self.sites:
                param_name = "sb{:d}{:d}_{:d}".format(x, y, c)
                D[param_name] = H[x + nn - 1, y - 1].real  # à modifier
            if self.complex:
                for y in self.sites:
                    if x == 1 and y == 1:
                        continue
                    param_name = "sb{:d}{:d}i_{:d}".format(x, y, c)
                    D[param_name] = H[x + nn - 1, y - 1].imag  # à modifier

        if self.triplet:
            for y in self.sites:
                param_name = "pb{:d}{:d}_{:d}".format(x, y, c)
                D[param_name] = H[x + nn - 1, y - 1].real  # à modifier
            if self.complex:
                for y in self.sites:
                    if x == 1 and y == 1:
                        continue
                    param_name = "pb{:d}{:d}i_{:d}".format(x, y, c)
                    D[param_name] = H[x + nn - 1, y - 1].imag  # à modifier
        return D

    # -----------------------------------------------------------------------------------------------
    def starting_values(
        self, c=1, e=(0.5, 1.5), hyb=(0.5, 0.2), shyb=(0.1, 0.05), pr=False
    ):
        """returns an initialization string for the bath parameters

        :param int c: cluster label (starts at 1)
        :param (float,float) e: bounds of the values for the bath energies (absolute value)
        :param (float,float) hyb: average and deviation of the normal hybridization parameters
        :param (float,float) shyb: average and deviation of the anomalous hybridization parameters
        :param bool pr: prints the starting values if True
        :return str: initialization string

        """
        S = ""
        fac = 1
        E = np.linspace(e[0], e[1], len(self.CM.var_E))
        for i, x in enumerate(self.CM.var_E):
            bn = int(x[2])
            fac = 2 * (bn % 2) - 1
            S += x + "_" + str(c) + " = " + str(fac * E[i]) + "\n"
        for x in self.CM.var_H:
            if x[0:2] == "sb" or x[0:2] == "pb":
                S += (
                    x
                    + "_"
                    + str(c)
                    + " = "
                    + str(shyb[0] + shyb[1] * (2 * np.random.random() - 1))
                    + "\n"
                )
            else:
                S += (
                    x
                    + "_"
                    + str(c)
                    + " = "
                    + str(hyb[0] + hyb[1] * (2 * np.random.random() - 1))
                    + "\n"
                )
        if pr:
            print("starting values:\n", S)
        return S

    # -----------------------------------------------------------------------------------------------
    def starting_values_PH(self, c=1, e=(1, 0.5), hyb=(0.5, 0.2), phi=None, pr=False):
        """returns an initialization string for the bath parameters, in the particle-hole symmetric case.

        :param int c: cluster label
        :param (float) e: range of bath energies
        :param (float,float) hyb: average and deviation of the normal hybridization parameters
        :param [int] phi: PH phases of the cluster sites proper
        :param bool pr: if True, prints info
        :return str: initialization string

        """
        S = ""
        fac = 1
        if self.nb % 2:
            raise ValueError(
                "the number of bath orbitals must be even for starting_values_PH() to apply"
            )
        if phi is None:
            raise ValueError("The PH phases of the sites must be specified")
        NE = self.nb // 2

        if self.spin_dependent or self.spin_flip:
            for i in range(NE):
                S += "eb{:d}u_{:d} = -1.0*eb{:d}d_{:d}\n".format(
                    2 * i + 2, c, 2 * i + 1, c
                )
                S += "eb{:d}d_{:d} = -1.0*eb{:d}u_{:d}\n".format(
                    2 * i + 2, c, 2 * i + 1, c
                )
                for s in self.sites[2 * i]:
                    S += "tb{:d}u{:d}u_{:d} = {:2.1f}*tb{:d}d{:d}d_{:d}\n".format(
                        2 * i + 2, s, c, phi[s - 1], 2 * i + 1, s, c
                    )
                    S += "tb{:d}d{:d}d_{:d} = {:2.1f}*tb{:d}u{:d}u_{:d}\n".format(
                        2 * i + 2, s, c, phi[s - 1], 2 * i + 1, s, c
                    )
                if self.complex:
                    for s in self.sites[2 * i]:
                        S += "tb{:d}u{:d}ui_{:d} = {:2.1f}*tb{:d}d{:d}di_{:d}\n".format(
                            2 * i + 2, s, c, -phi[s - 1], 2 * i + 1, s, c
                        )
                        S += "tb{:d}d{:d}di_{:d} = {:2.1f}*tb{:d}u{:d}ui_{:d}\n".format(
                            2 * i + 2, s, c, -phi[s - 1], 2 * i + 1, s, c
                        )
                if self.spin_flip:
                    for s in self.sites[2 * i]:
                        S += "tb{:d}u{:d}d_{:d} = {:2.1f}*tb{:d}u{:d}d_{:d}\n".format(
                            2 * i + 2, s, c, -phi[s - 1], 2 * i + 1, s, c
                        )
                        S += "tb{:d}d{:d}u_{:d} = {:2.1f}*tb{:d}d{:d}u_{:d}\n".format(
                            2 * i + 2, s, c, -phi[s - 1], 2 * i + 1, s, c
                        )
                    if self.complex:
                        for s in self.sites[2 * i]:
                            S += "tb{:d}u{:d}di_{:d} = {:2.1f}*tb{:d}u{:d}di_{:d}\n".format(
                                2 * i + 2, s, c, phi[s - 1], 2 * i + 1, s, c
                            )
                            S += "tb{:d}d{:d}ui_{:d} = {:2.1f}*tb{:d}d{:d}ui_{:d}\n".format(
                                2 * i + 2, s, c, phi[s - 1], 2 * i + 1, s, c
                            )

        else:
            for i in range(NE):
                S += "eb{:d}_{:d} = -1.0*eb{:d}_{:d}\n".format(
                    2 * i + 2, c, 2 * i + 1, c
                )
                for s in self.sites[2 * i]:
                    S += "tb{:d}{:d}_{:d} = {:2.1f}*tb{:d}{:d}_{:d}\n".format(
                        2 * i + 2, s, c, phi[s - 1], 2 * i + 1, s, c
                    )
                if self.complex:
                    for s in self.sites[2 * i]:
                        S += "tb{:d}{:d}i_{:d} = {:2.1f}*tb{:d}{:d}i_{:d}\n".format(
                            2 * i + 2, s, c, -phi[s - 1], 2 * i + 1, s, c
                        )

        var_E = self.CM.var_E
        self.CM.var_E = []
        for x in var_E:
            for i in range(NE):
                if "eb{:d}".format(2 * i + 1) in x:
                    self.CM.var_E.append(x)
        var_H = self.CM.var_H
        self.CM.var_H = []
        for x in var_H:
            for i in range(NE):
                if "tb{:d}".format(2 * i + 1) in x:
                    self.CM.var_H.append(x)

        for i, x in enumerate(self.CM.var_E):
            S += (
                x
                + "_"
                + str(c)
                + " = "
                + str(e[0] + e[1] * (2 * np.random.random() - 1))
                + "\n"
            )
        for x in self.CM.var_H:
            S += (
                x
                + "_"
                + str(c)
                + " = "
                + str(hyb[0] + hyb[1] * (2 * np.random.random() - 1))
                + "\n"
            )
        if pr:
            print("starting values:\n", S)
        return S


######################################################################
class hybridization:
    """
    Defines hybridization data from a data file.
    the frequency is purely imaginary; it is its imaginary part that appears in the file

    :param str file: name of the file or string. Format : each line starts with a frequency and then has N*N columns for Delta_{ij}(w)

    """

    def __init__(self, file):
        NN = [1, 4, 9, 16, 25, 36, 49, 64, 81, 100, 121, 144]
        # data = np.genfromtxt(file, dtype=np.complex128)
        # data = np.genfromtxt(file, invalid_raise=True, loose=False, dtype=None)
        data = np.genfromtxt(file, invalid_raise=True, loose=False, dtype=np.complex128)
        # print(data)
        self.nw = data.shape[0]
        LL = data.shape[1] - 1
        try:
            self.n = NN.index(LL) + 1
        except ValueError:
            raise ValueError(
                "The number of columns in the hybridization table should be 1 + n^2"
            ) from None

        self.w = np.copy(np.real(data[:, 0]))
        self.Delta = np.zeros((self.nw, self.n, self.n), dtype=np.complex128)
        for i in range(self.nw):
            self.Delta[i, :, :] = np.reshape(data[i, 1:], (self.n, self.n))

    # -----------------------------------------------------------------------------------------------
    def distance(self, I):
        """
        Evaluates the distance function bewteen the data and the hybridization function of an impurity model instance of label I

        :param [model_instance] I: model instance
        :returns: the squared Frobenius norm of the difference matrix, averaged over frequencies
        :rtype: float
        """
        M = np.zeros((self.nw, self.n, self.n), dtype=np.complex128)
        for i in range(self.nw):
            M[i, :, :] = I.hybridization_function(self.w[i] * 1j, spin_down=False)

        # print('M : '); print(M)
        # print('Delta : '); print(self.Delta)

        dist = np.linalg.norm(M - self.Delta)
        return dist * dist / self.nw

    # -----------------------------------------------------------------------------------------------
    def __str__(self):
        S = ""
        for i in range(self.nw):
            S += "w = " + self.w[i].__str__() + "\n" + self.Delta[i].__str__() + "\n\n"
        return S

    # -----------------------------------------------------------------------------------------------
    def optimize_bath(
        self, model, varia, x, method="Nelder-Mead", accur=1e-4, accur_dist=1e-8
    ):
        """
        Optimizes the bath parameters to fit the hybridization function delta
        :param lattice_model model: the lattice model whose bath parameters are to be optimized
        :param [str] varia: list of variational parameters (bath parameters) from PyQCM
        :param [float] x: starting values of the parameters
        :param str method: minimization method passed to optimize() (default 'Nelder-Mead')
        :param float accur: requested accuracy in the parameters (x-tolerance)
        :param float accur_dist: requested accuracy in the distance function (f-tolerance)
        """

        nvar = len(varia)

        def tmp_distance(x):
            for i in range(nvar):
                model.set_parameter(varia[i], x[i])
            I = pyqcm.model_instance(model)
            return self.distance(I)

        return optimize(
            tmp_distance, x, method=method, accur=accur, accur_dist=accur_dist
        )


####################################################################################################
# PRIVATE FUNCTIONS


# ---------------------------------------------------------------------------------------------------
def check_bounds(x, B=100, v=None):
    """Checks whether one of the variables in the array x is out of bounds

    :param [float] x: array of parameters
    :param float B: maximum value of all parameters
    :param [str] v: list of corresponding names, if available
    :returns:None
    """
    for i in range(len(x)):
        if np.abs(x[i]) > B:
            if v != None:
                S = "parameter {:s} = {:g} is out of bounds!".format(v[i], x[i])
            else:
                S = "parameter no {:d}  = {:g} is out of bounds!".format(i + 1, x[i])
            print(
                "Try setting `max_value` to a bigger value within the CDMFT procedure!"
            )
            raise pyqcm.OutOfBoundsError(S)


# ---------------------------------------------------------------------------------------------------
# optimization of the bath parameters


def optimize(
    F,
    x,
    method="Nelder-Mead",
    initial_step=0.1,
    accur=1e-4,
    accur_dist=1e-8,
    maxfev=5000000,
):
    """

    :param F: function to be optimized
    :param (float) x: array of variables
    :param str method: method to use, as used in scipy.optimize.minimize()
    :param float initial_step: initial step in the minimization procedure
    :param float accur: requested accuracy in the parameters
    :param float accur_dist: requested accuracy in the distance function
    :param int maxfev: maximum number of function evaluations allowed
    :returns: a 4-tuple (opt_x, iter_done, success, fun) with the optimal parameters, number of evaluations, success flag, and optimal function value
    :rtype: tuple

    """

    displaymin, nvar = False, len(x)
    nlopt_methods = {
        "COBYLA": nlopt.LN_COBYLA,
        "BOBYQA": nlopt.LN_BOBYQA,
        "PRAXIS": nlopt.LN_PRAXIS,
        "NELDERMEAD": nlopt.LN_NELDERMEAD,
        "SUBPLEX": nlopt.LN_SBPLX,
    }
    if method == "Nelder-Mead":
        initial_simplex = np.zeros((nvar + 1, nvar))
        for i in range(nvar + 1):
            initial_simplex[i, :] = x
        for i in range(nvar):
            initial_simplex[i + 1, i] += initial_step
        sol = scipy.optimize.minimize(
            F,
            x,
            method="Nelder-Mead",
            options={
                "disp": displaymin,
                "maxfev": maxfev,
                "xatol": accur,
                "fatol": accur_dist,
                "initial_simplex": initial_simplex,
                "adaptive": True,
                "disp": False,
            },
        )
        opt_x, iter_done, success, fun = sol.x, sol.nfev, sol.success, sol.fun

    elif method == "Powell":
        sol = scipy.optimize.minimize(F, x, method="Powell", tol=accur)
        opt_x, iter_done, success, fun = sol.x, sol.nfev, sol.success, sol.fun

    elif method == "CG":
        sol = scipy.optimize.minimize(
            F, x, method="CG", jac=False, tol=accur, options={"gtol": 1e-6, "eps": 1e-6}
        )
        opt_x, iter_done, success, fun = sol.x, sol.nfev, sol.success, sol.fun

    elif method == "BFGS":
        sol = scipy.optimize.minimize(
            F, x, method="BFGS", jac=False, tol=accur, options={"eps": accur}
        )
        opt_x, iter_done, success, fun = sol.x, sol.nfev, sol.success, sol.fun

    elif method == "ANNEAL":
        sol = scipy.optimize.basinhopping(
            F,
            x,
            minimizer_kwargs={
                "method": "COBYLA",
                "options": {
                    "disp": displaymin,
                    "rhobeg": initial_step,
                    "maxiter": maxfev,
                    "rtol": accur_dist,
                },
            },
        )
        opt_x, iter_done, success, fun = sol.x, sol.nfev, sol.success, sol.fun

    elif method in nlopt_methods.keys():
        optimizer = nlopt.opt(nlopt_methods[method], nvar)

        # Set objective function and parameters bounds (same as in `check_bounds` method)
        optimizer.set_min_objective(F)
        optimizer.set_lower_bounds(np.array([-np.inf for _ in x]))
        optimizer.set_upper_bounds(np.array([np.inf for _ in x]))

        # Set function values (ftol) and parameters (xtol) tolerances
        optimizer.set_ftol_rel(accur_dist)
        optimizer.set_xtol_rel(accur)

        optimizer.set_maxeval(maxfev)
        optimizer.set_initial_step(initial_step)

        # Minimizing...
        sol = optimizer.optimize(x)
        opt_x, iter_done, success, fun = (
            sol,
            optimizer.get_numevals(),
            optimizer.last_optimize_result(),
            optimizer.last_optimum_value(),
        )

    else:
        raise ValueError(f"unknown method specified for minimization: {method}")

    if not success:
        raise pyqcm.MinimizationError(
            "Failure of the optimization of the bath parameters"
        )

    return opt_x, iter_done, success, fun


if __name__ == "__main__":
    pass



import pyqcm
import numpy as np

def model1D(L, sym=True):
    if sym:
        g = [[L-i for i in range(L)]]
        print('generators :', g)
        CM = pyqcm.cluster_model(L, 0, generators=g)
    else:
        CM = pyqcm.cluster_model(L, 0)
    pos = [(i,0,0) for i in range(L)]
    clus = pyqcm.cluster(CM, pos)
    model = pyqcm.lattice_model('1D_{:d}'.format(L), clus, ((L, 0, 0),))

    model.interaction_operator('U')
    model.interaction_operator('V', link=( 1, 0, 0))
    model.hopping_operator('t', ( 1, 0, 0), -1)
    model.hopping_operator('tp', ( 2, 0, 0), -1)
    model.hopping_operator('hx', [0, 0, 0], 1, tau=0, sigma=1)
    model.anomalous_operator('D', ( 1, 0, 0), 1)

    if (L%2 == 0) and sym == False:
        model.density_wave('cdw', 'N', ( 1, 0, 0))
        e = np.sqrt(0.5)
        model.explicit_operator('V0m', (
            ((0,0,0), (0,0,0), e),
            ((L-1,0,0), (0,0,0), e),
        ), tau=0, type='one-body')
        model.explicit_operator('V1m', (
            ((0,0,0), (0,0,0), e),
            ((L-1,0,0), (0,0,0),-e),
        ), tau=0, type='one-body')

        model.explicit_operator('Vf', (((L-1,0,0),(1,0,0),1),))
        model.V0m_eig = 1
        model.V1m_eig = -1
        model.Vf_eig = 1/8

    return model




################################################################################
# This file contains functions implementing the Variational Cluster Approximation (VCA)
################################################################################

import numpy as np
import pyqcm
import time
import nlopt

from scipy.optimize import minimize

#-------------------------------------------------------------------------------
# MPI matter
comm = None
root = False
try:
    from mpi4py import MPI
    comm = MPI.COMM_WORLD
    if comm.Get_rank() == 0:
        root = True
        print('MPI available. Working on ', comm.Get_size(), ' concurrent processes')
except ImportError:
    root = True

current_instance = None
verbose = True

################################################################################
# PRIVATE FUNCTIONS
################################################################################


################################################################################
# evaluation of many instances of a function in parallel

def _evalF(func, x):
    n = x.shape[0]
    y = np.empty(n)
    for i in range(n):
        y[i] = func(x[i,:])
    return y    

def _evalF_mpi(func, x):
    n = x.shape[0]
    nproc = comm.Get_size()
    rank = comm.Get_rank()
    y = np.zeros(n)
    y_tmp = np.zeros(n)
    k = 0  # index of points
    for i in range(n):
        if i % nproc == rank:
            y_tmp[i] = func(x[i,:])

    comm.Allreduce(y_tmp, y)
    return y


################################################################################
# quasi-Newton method function evaluation

def _QN_values(func, x, step):
    """Computes the function func of n variables at :math:`2n+1` points at x and at n pairs of points located at +/- step from x in each direction. For use in the quasi-Newton method.

    :param func: function of n variables
    :param x: base point (n components NumPy array)
    :param float step: step to take in each direction in computing derivatives
    :returns: array of :math:`2n+1` values

    """
    n = len(x)
    N = 2 * n + 1
    xp = np.empty((N, n))
    for i in range(N):
        xp[i, :] = np.copy(x)
    k = 0
    for i in range(n):
        k += 1
        xp[k, i] += step[i]
        k += 1
        xp[k, i] -= step[i]

    if comm is None:
        return _evalF(func, xp)
    else:    
        return _evalF_mpi(func, xp)


################################################################################
# quasi-Newton method

def _quasi_newton(func=None, start=None, step=None, accur=None, max=10, gtol=1e-4, bfgs=False, max_iteration=30, max_iter_diff=None, hartree=None):
    """Performs the quasi_newton procedure
    
    :param func: a function of N variables
    :param [float] start: the starting values
    :param [float] step: the steps used to computed the numerical second derivatives
    :param [float] accur: the required accuracy for each variable
    :param [float] max: maximum absolute value of each parameter
    :param float gtol: the gradient tolerance (gradient must be smaller than gtol for convergence)
    :param bool bfgs: True if the BFGS method is used, otherwise the symmetric rank-1 formula is used (default)
    :param int max_iterations: maximum number of iterations, beyond which an exception is raised
    :param float max_iter_diff: maximum step to make
    :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
    :returns: tuple of x (the solution), gradient (array, the value of the gradient), hessian (matrix, the Hessian matrix)
    :rtype: tuple[float, list[float], list[list[float]]]

    """
    global root, verbose

    n = len(start)
    gradient = np.zeros(n)
    gradient0 = np.zeros(n)
    dx = np.zeros(n)
    y = np.zeros(n)
    ihessian = np.eye(n)  # inverse Hessian matrix
    iteration = 0
    x = start
    hartree_converged = True
    step0 = step

    while iteration < max_iteration:
        x0 = x
        x, dx, gradient, ihessian = _quasi_newton_step(iteration, func, x0, step, gradient, dx, bfgs)
        if root: _vca_accuracy_warning(accur, ihessian)
        iteration += 1

        if hartree != None:
            hartree_converged = True
            for C in hartree:
                C.update(current_instance)
                if root and verbose: C.print()
                hartree_converged = hartree_converged and C.converged()

        if (np.linalg.norm(gradient) < gtol) and hartree_converged :
            if root and verbose:
                print('convergence on gradient after ', iteration, ' iterations')
            break

        converged = True
        for i in range(n):
            if np.abs(dx[i]) > accur[i]:
                converged = False
                break
        if converged and hartree_converged:
            if root and verbose:
                print('convergence on position after ', iteration, ' iterations')
            break

        #redefining the steps
        step_multiplier = 2.0
        for i in range(n):
            step[i] = np.abs(dx[i])
            if step[i] > step0[i]:
                step[i] = 2.0*step0[i]
            if step[i] < step_multiplier*accur[i]:
                step[i] = step_multiplier*accur[i]

    if iteration == max_iteration:
        raise pyqcm.TooManyIterationsError(max_iteration)

    return x, gradient, ihessian

################################################################################
# quasi-Newton method (step)

def _quasi_newton_step(iteration = 0, func=None, x=None, step=None, gradient=None, dx=None, bfgs=False, max_diff=None):
    """Performs a step of the quasi_newton procedure
    
    :param int iter: iteration number
    :param func: a function of N variables
    :param [float] x: the starting values
    :param [float] step: the steps used to computed the numerical second derivatives
    :param [float] gradient: the gradient array from the previous step (updated in place)
    :param [float] dx: the previous difference
    :param bool bfgs: True if the BFGS method is used, otherwise the symmetric rank-1 formula is used (default)
    :param float max_diff: maximum step to make
    :returns: tuple of x (the new point), dx (the difference between the current and previous point), gradient (array, the value of the gradient), hessian (matrix, the Hessian matrix)
    :rtype: tuple[float, float, list[float], list[list[float]]]

    """
    global verbose

    n = len(x)
    y = np.zeros(n)
    ihessian = np.eye(n)  # inverse Hessian matrix

    F = _QN_values(func, x, step)
    gradient0 = np.copy(gradient)
    for i in range(n):
        gradient[i] = (F[2 * i + 1] - F[2 * i + 2]) / (2 * step[i])

    if iteration == 0:
        for i in range(n):
            ihessian[i, i] = step[i] * step[i] / (F[2 * i + 1] + F[2 * i + 2] - 2 * F[0])

    else:
        y = gradient - gradient0
        if bfgs:  # BFGS method
            z1 = 1.0e-14 + np.dot(y, dx)
            z1 = 1.0 / z1
            X = np.eye(n)
            X -= z1 * np.kron(y, dx).reshape((n, n))
            ihessian = np.dot(X.T, np.dot(ihessian, X))
            ihessian += z1 * np.kron(dx, dx).reshape((n, n))
        else:  # symmetric rank 1
            u = np.dot(ihessian, y)  # $u = H y_k$
            dx -= u  # the $\Delta x_k - H_k y_k$ of Wikipedia
            z1 = 1.0e-14 + np.dot(y, dx)
            z1 = 1.0 / z1
            ihessian += z1 * np.kron(dx, dx).reshape((n, n))

    dx = -np.dot(ihessian, gradient)

    if max_diff is not None:
        norm_dx = np.linalg.norm(dx)
        if norm_dx > max_diff:
            ratio = max_diff / norm_dx
            dx *= ratio
            if root:
                s1 = 'Norm of QN iteration variation dx is out of bounds: {} > {}'.format(norm_dx, max_diff)
                s2 = 'Readjusting: dx = dx * {:.4}%'.format(ratio * 100)
                if verbose: print(s1, '\n', s2)

    x += dx
    if root and verbose: print('QN iteration no ', iteration+1)

    return x, dx, gradient, ihessian


################################################################################
# Newton-Raphson method :  computation of the Hessian

def _NR_Hessian(func, x, step):
    """Computes the function func of n variables at :math:`1 + 2n + n(n-1)/2` points in order to fit a quadratic form x and n pairs of points located at +/- step from x in each direction
    
    :param func: the function
    :param [float] x: the base point 
    :param float step: step taken in each direction
    :returns: the value of the function, the gradient, and the Hessian
    :rtype: tuple[float, list[float], list[list[float]]]

    """
    n = len(x)
    N = 1 + 2*n + (n*(n-1))//2
    y = np.zeros(N)
    hess = np.zeros((n,n))

    xp = np.zeros((N, n))
    for k in range(N):
        xp[k, :] = np.copy(x)
 
    k = 0
    # points located at +/- step
    for i in range(n):
        k += 1
        xp[k, i] += step[i]
        k += 1
        xp[k, i] -= step[i]

    # points located off diagonal
    for i in range(n):
        for j in range(i):
            k += 1
            xp[k, i] += 0.707*(2*(i%2)-1)*step[i]
            xp[k, j] += 0.707*(2*(j%2)-1)*step[j]

    if comm is None:
        y = _evalF(func, xp)
    else:    
        y = _evalF_mpi(func, xp)

    # setting fit coefficients
    A = np.empty((N,N))
    A[:, 1:n+1] = xp
    for k in range(N):
        A[k,0] = 1.0
        m = n+1
        for i in range(n):
            for j in range(i+1):
                A[k, m] = xp[k, i]*xp[k,j]
                if i == j :
                    A[k, m] = xp[k, i]*xp[k,j]*0.5
                else:
                    A[k, m] = xp[k, i]*xp[k,j]        
                m += 1

    # inverting
    A1 = np.linalg.inv(A)
    y = np.dot(A1, y)
    val = y[0]
    grad = y[1:n+1]
    val += np.sum(grad*x)
    m = n+1
    for i in range(n):
        for j in range(i+1):
            hess[i,j] = y[m]
            hess[j,i] = y[m]
            if i == j :
                val += x[i] * x[j] * y[m] * 0.5
            else:
                val += x[i] * x[j] * y[m]
            m += 1
    grad += np.dot(hess,x)
    return val, grad, hess


################################################################################
# Newton-Raphson method

def _newton_raphson(func=None, start=None, step=None, accur=None, max=10, gtol=1e-4, max_iteration=30, max_iter_diff=None, hartree=None):
    """Performs the Newton-Raphson procedure
    
    :param func: a function of N variables
    :param [float] start: the starting values
    :param [float] step: the steps used to computed the numerical second derivatives
    :param [float] accur: the required accuracy for each variable
    :param [float] max: maximum absolute value of each parameter
    :param float gtol: the gradient tolerance (gradient must be smaller than gtol for convergence)
    :param int max_iterations:  maximum number of iterations, beyond which an exception is raised
    :param float max_iter_diff: optional maximum value of the maximum step
    :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
    :returns: the value of the function, the gradient, and the Hessian
    :rtype: tuple[float, list[float], list[list[float]]]

    """
    global current_instance, root, verbose

    n = len(start)
    gradient = np.zeros(n)
    dx = np.zeros(n)
    y = np.zeros(n)
    iteration = 0
    x = start
    step0 = step
    hartree_converged = True

    while iteration < max_iteration:
        iteration += 1

        gradient0 = np.copy(gradient)
        dx, F, gradient, hessian = _newton_raphson_step(func, x, step)
        ihessian = np.linalg.inv(hessian)
        _vca_accuracy_warning(accur, ihessian)

        if hartree != None:
            hartree_converged = True
            for C in hartree:
                C.update(current_instance, pr=True)
                hartree_converged  = hartree_converged and C.converged()

        if np.linalg.norm(gradient) < gtol and hartree_converged:
            if root and verbose:
                print('convergence on gradient after ', iteration, ' iterations')
            break


        if max_iter_diff is not None:
            norm_dx = np.linalg.norm(dx)
            if norm_dx > max_iter_diff:
                ratio = max_iter_diff / norm_dx
                dx *= ratio
                s1 = 'Norm of NR iteration variation dx is out of bounds: {} > {}'.format(norm_dx, max_iter_diff)
                s2 = 'Readjusting: dx = dx * {:.4}%'.format(ratio * 100)
                if root and verbose:
                    print(s1, '\n', s2)
                    
        x += dx

        #redefining the steps
        step_multiplier = 2.0
        for i in range(n):
            step[i] = np.abs(dx[i])
            if step[i] > step0[i]:
                step[i] = 2.0*step0[i]
            if step[i] < step_multiplier*accur[i]:
                step[i] = step_multiplier*accur[i]

        if root and verbose:
            print('NR iteration no ', iteration, '\t x = ', x, '\t steps = ', step, '\t dx = ', dx)

        converged = True
        for i in range(n):
            if np.abs(dx[i]) > accur[i]:
                converged = False
                break
        if converged and hartree_converged:
            if root and verbose:
                print('convergence on position after ', iteration, ' iterations')
            break

    if iteration == max_iteration:
        raise pyqcm.TooManyIterationsError(max_iteration)

    return x, gradient, ihessian

################################################################################
# Newton-Raphson step

def _newton_raphson_step(func=None, x=None, step=None):
    """Performs a Newton-Raphson step
    
    :param func: a function of N variables
    :param [float] x: the current value of the parameters (changed by the function)
    :param [float] step: the steps used to computed the numerical second derivatives
    :returns: the change in position, the value of the function, the gradient, and the Hessian
    :rtype: tuple[list[float], float, list[float], list[list[float]]]

    """

    F, gradient, hessian = _NR_Hessian(func, x, step)
    ihessian = np.linalg.inv(hessian)
    dx = -np.dot(ihessian, gradient)

    return dx, F, gradient, ihessian


################################################################################
# checking consistency of accuracy with effective precision

def _vca_accuracy_warning(accur, ihessian):
    global verbose

    accur_SEF = pyqcm.get_global_parameter('accur_SEF')
    inv_der2 = np.diagonal(ihessian)
    inv_der2 = np.sqrt(accur_SEF*np.abs(inv_der2))
    if verbose: print('effective VCA precision: ', inv_der2)
    for i in range(len(accur)):
        if inv_der2[i] > accur[i]:
            print('WARNING : effective accuracy of variable {:d} lower than what is required. Maybe lower required accuracy (i.e. increase "accur") or lower "accur_SEF"'.format(i+1))
    

################################################################################
# Newton-Raphson in 1D

def _altNR(func=None, start=None, step=None, accur=None, max=10, gtol=1e-4, max_iteration=30, max_iter_diff=None, hartree=None):
    """Performs a Newton-Raphson procedure in one variable by reusing old points as much as possible
    
    :param func: a function of 1 variable
    :param [float] start: the starting value
    :param [float] step: the first step 
    :param [float] accur: the required accuracy
    :param [float] max: maximum absolute value of the parameter
    :param float gtol: the gradient tolerance (gradient must be smaller than gtol for convergence)
    :param int max_iterations:  maximum number of iterations, beyond which an exception is raised
    :param float max_iter_diff: optional maximum value of the maximum step
    :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
    :returns: the value of the function, the gradient, and the 2nd derivative
    :rtype: tuple[float, float, float]

    """
    global verbose

    # initial three points
    X = np.zeros(3)
    Y = np.zeros(3)
    X[0] = start[0]-step[0]
    X[1] = start[0]
    X[2] = start[0]+step[0]
    Y[0] = func([X[0]])
    Y[1] = func([X[1]])
    Y[2] = func([X[2]])

    iter = 0
    xp0 = 1e9
    accur_SEF = pyqcm.get_global_parameter('accur_SEF')
    while iter < max_iteration:
        iter += 1
        C = np.polyfit(X,Y,2)
        xp = -0.5*C[1]/C[0]

        der1 = 2*C[0]*xp0 + C[1]
        der2 = 2*C[0]
        dx = np.sqrt(accur_SEF/np.abs(der2))
        if iter > 1 and verbose:
            print('---> X :', X, '\tder1 = {:.6g}\tder2 = {:.3g}'.format(der1, der2), '\teffective precision: {:.3g}'.format(dx))
            if dx > accur:
                print('WARNING : effective accuracy lower than what is required. Maybe lower required accuracy (i.e. increase "accur") or lower "accur_SEF"')
        if np.abs(der1) < gtol:
            if verbose: print('VCA 1D : convergence on gradient')
            break
        I = np.argmax(np.abs(X-xp))
        X[I] = xp
        Y[I] = func([X[I]])
        if iter > 1 and np.abs(xp-xp0) < accur:
            if verbose: print('VCA 1D : convergence on position')
            break
        xp0 = xp
    if iter == max_iteration:
        raise pyqcm.TooManyIterationsError(max_iteration)

    return np.array([xp]), np.array([der1]), np.array([[1.0/der2]])


####################################################################################################
# performs the VCA
SEF_eval = 0

class VCA:
    """
    class containing the elements of a VCA computation. The constructor executes the computation.

    :param lattice_model model: the (unique) model on which the computation is based
    :param var2sef: function that converts variational parameters to model parameters
    :param str or [str] varia: variational parameters (list or tuple)
    :param float or [float] start: starting values
    :param float or [float] steps: initial steps
    :param float or [float] accur: accuracy of parameters (also step for 2nd derivatives)
    :param float or [float] max: maximum values that are tolerated
    :param float accur_grad: max value of gradient for convergence
    :param int max_iter: maximum number of iterations in the procedure
    :param float max_iter_diff: optional maximum value of the maximum step in the quasi-Newton method
    :param str method: method used to optimize ('SYMR1', 'NR', 'BFGS', 'altNR', 'Nelder-Mead', 'COBYLA', 'Powell', 'CG', 'minimax')
    :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
    :param bool hartree_self_consistent: True if the Hartree approximation is treated in the self-consistent, rather than variational, way.
    :param str symmetrized_operator: name of an operator wrt which the functional must be symmetrized
    :param int var_max_start: label of the first variable for which the function is a maximum (minimal vars first, maximal vars last)
    :param bool verb: if True (default) prints ample progress messages
    :returns:None

    :ivar lattice_model model: (unique) model on which the computation is based
    :ivar I: current model instance (changes in the course of the computation)
    :ivar hessian: Hessian matrix of the Potthoff functional (matrix of second derivatives)
    
    """
    first_time = True
    def __init__(self, model,
        var2sef=None,
        varia=None, 
        start=None, 
        steps=0.01, 
        accur=1e-4, 
        max=100,
        file="vca.tsv",
        accur_grad=1e-6, 
        max_iter=30, 
        max_iter_diff=None, 
        method='NR', 
        hartree=None, 
        hartree_self_consistent=False,
        symmetrized_operator=None,
        var_max_start = None,
        consistency_check=False,
        verb = True
    ):
        global verbose
        self.model = model
        verbose = verb
        global SEF_eval
        # type and length checks
        if pyqcm.is_sequence(varia) == False:
            if type(varia) != str:
                raise ValueError('argument varia of vca() must be a string or a sequence of strings')
            varia = (varia,)

        if isinstance(accur_grad, float) == False: raise ValueError("'accur_grad' in VCA() should be a single float")
        
        nvar = len(varia)
        if len(set(varia)) != nvar: raise ValueError('There are duplicate elements in the list of Weiss fields')
        if type(steps) != list: 
            if type(steps) == float : steps = [steps]*nvar
            else : steps = list(steps)
        if pyqcm.is_sequence(accur) == False: accur = [accur]*nvar
        if type(accur) == tuple: accur = list(accur)
        if pyqcm.is_sequence(max) == False: max = (max,)*nvar
        if pyqcm.is_sequence(hartree) == False and hartree is not None: hartree = (hartree,)

        self.I = pyqcm.model_instance(model)
        L = model.nsites

        hartree_self = None
        if hartree_self_consistent:
            hartree_self = hartree

        if varia is None:
            print('missing argument varia : variational parameters must be specified')
            raise pyqcm.MissingArgError('varia')

        if max is None:
            max = 10*np.ones(nvar)
        elif len(max) != nvar:
            print('the argument "max" should have ', nvar, ' components')
            raise pyqcm.MissingArgError('max')

        if start is None:
            if var2sef != None:
                raise ValueError('the start argument is missing in vca(). Should be specified if var2sef is not None')
            start = [0.0]*nvar
            P = model.parameters()
            for i,v in enumerate(varia):
                start[i] = P[v]

        if pyqcm.is_sequence(start) == False: start = (start,)*nvar
        elif len(start) != nvar:
            print('the argument "start" should have ', nvar, ' components')
            raise pyqcm.MissingArgError('start')
        start = np.array(start)

        if accur is None:
            accur = 1e-4*np.ones(nvar)

        if steps is None:
            steps = 10*np.array(accur)
        elif len(steps) != nvar:
            print('the argument "steps" should have ', nvar, ' components')
            raise pyqcm.MissingArgError('steps')
    
        self.consistency_check = consistency_check

        SEF_eval = 0
        def var2x(x, grad=None):
            global current_instance, root
            for i in range(len(x)):
                if np.abs(x[i]) > max[i]:
                    raise pyqcm.OutOfBoundsError('parameter {:s} is out of bounds!'.format(varia[i]))
            global SEF_eval
            if var2sef is None:
                for i in range(len(varia)): 
                    model.set_parameter(varia[i], x[i])
                if verbose: print('x = ', x) # new
            else:
                var2sef(x)    
            SEF_eval += 1
            self.I = pyqcm.model_instance(model)
            current_instance = self.I
            return self.I.Potthoff_functional(hartree, symmetrized_operator=symmetrized_operator,consistency_check=self.consistency_check)
            
        if root:
            if hartree is None:
                pyqcm.banner('VCA procedure, method {:s}'.format(method), '*')
            else:
                pyqcm.banner('VCA procedure, (combined with Hartree procedure) method {:s}'.format(method), '*')
            var_val = pyqcm.varia_table(varia,start)
            if verbose: print(var_val)

        nlopt_methods = {
            'COBYLA': nlopt.LN_COBYLA,
            'BOBYQA': nlopt.LN_BOBYQA,
            'PRAXIS': nlopt.LN_PRAXIS,
            'NELDERMEAD': nlopt.LN_NELDERMEAD,
            'SUBPLEX': nlopt.LN_SBPLX
        }

        lib_minimization = False
        if method == 'Nelder-Mead' or method == 'COBYLA' or method == 'Powell'  or method == 'CG'  or method == 'BFGS'  or method == 'minimax' or method in nlopt_methods.keys():
            lib_minimization = True
        if lib_minimization and hartree_self_consistent:
            raise ValueError('Hartree self-consistency not allowed when using SciPy or nlopt minimization methods in VCA')

        ftol = 2*pyqcm.qcm.get_global_parameter('accur_SEF')
        iH = None

        #..................................... SWITCH ACCORDING TO METHOD .........................................
        maxfev = 30*nvar*max_iter
        try:
            if method == 'altNR':
                if nvar != 1:
                    raise ValueError('vca method altNR only works with one variable')
                sol, grad, iH = _altNR(var2x, start, steps, accur, max, accur_grad, max_iteration=max_iter, max_iter_diff=max_iter_diff, hartree=hartree_self)  # special Newton-Raphson process in 1 variable

            elif method == 'NR' :
                sol, grad, iH = _newton_raphson(var2x, start, steps, accur, max, accur_grad, max_iteration=max_iter, max_iter_diff=max_iter_diff, hartree=hartree_self)  # Newton-Raphson process

            elif method == 'SYMR1' :
                sol, grad, iH = _quasi_newton(var2x, start, steps, accur, max, accur_grad, False, max_iteration=max_iter, max_iter_diff=max_iter_diff, hartree=hartree_self)  # quasi-Newton process

            elif method == 'BFGS2' :
                sol, grad, iH = _quasi_newton(var2x, start, steps, accur, max, accur_grad, True, max_iteration=max_iter, max_iter_diff=max_iter_diff, hartree=hartree_self)  # quasi-Newton process

            elif method == 'Nelder-Mead':
                initial_simplex = np.zeros((nvar+1,nvar))
                for i in range(nvar+1):
                    initial_simplex[i, :] = start
                for i in range(nvar):
                    initial_simplex[i+1, i] += steps[i]
                if type(accur) == list:
                    accur = accur[0]
                solution = minimize(var2x, start, method='Nelder-Mead', options={'maxfev':maxfev, 'xatol': accur, 'fatol':ftol, 'initial_simplex': initial_simplex, 'adaptive': True, 'disp':True})
                iter_done = solution.nit
                sol = solution.x

            elif method == 'Powell':
                accur = accur[0]
                solution = minimize(var2x, start, method='Powell', tol = ftol, options={'xtol': accur, 'ftol': ftol, 'disp':True})
                iter_done = solution.nit
                sol = solution.x
                if type(sol) != list:
                    sol = [sol]

            elif method == 'CG':
                solution = minimize(var2x, start, method='CG', jac=False, tol = ftol, options={'eps':accur, 'maxiter':3*max_iter, 'disp':True})
                iter_done = solution.nit
                sol = solution.x

            elif method == 'BFGS':
                solution = minimize(var2x, start, method='BFGS', jac=False, tol = ftol, options={'eps':accur, 'disp':True})
                iter_done = solution.nit
                sol = solution.x
                iH = solution.hess_inv

            elif method == 'COBYLA':
                solution = minimize(var2x, start, method='COBYLA', options={'rhobeg':steps[0], 'maxiter':30*nvar*max_iter, 'tol': ftol, 'disp':True})
                iter_done = solution.nfev
                sol = solution.x

            elif method == 'minimax':
                sol = self._minimax(varia, var_max_start, start, steps, accur, max, max_iteration=max_iter, hartree=hartree)

            elif method in nlopt_methods.keys():
                optimizer = nlopt.opt(nlopt_methods[method], len(start))

                # Set objective function and parameters bounds (same as in `check_bounds` method)
                optimizer.set_min_objective(var2x)
                optimizer.set_lower_bounds(np.array([-np.inf for _ in start]))
                optimizer.set_upper_bounds(np.array([np.inf for _ in start]))

                # Set function values (ftol) and parameters (xtol) tolerances
                optimizer.set_ftol_rel(ftol)
                optimizer.set_xtol_rel(accur[0])

                optimizer.set_maxeval(maxfev)
                optimizer.set_initial_step(steps[0])

                # Minimizing...
                sol = optimizer.optimize(start)
                # iter_done = optimizer.get_numevals()
                # success = optimizer.last_optimize_result(), optimizer.last_optimum_value()
                # fun = optimizer.last_optimum_value()

            else:
                raise ValueError('method {:s} unknown in VCA'.format(method))

        #..........................................................................................................

        except pyqcm.OutOfBoundsError as E:
            raise pyqcm.SolverError('Failure of the VCA method') from E

        except pyqcm.TooManyIterationsError as E:
            raise pyqcm.SolverError('Failure of the VCA method') from E

        omega = var2x(sol)  # final, converged value
        if root:
            if iH is not None: 
                H = np.linalg.inv(iH)  # Hessian at the solution (inverse of iH)
            if verbose:
                print('saddle point = ', sol)
                if lib_minimization is False: 
                    print('gradient = ', grad)
                    print('second derivatives :', np.diag(H))
                    print('eigenvalues of Hessian :', np.linalg.eigh(H)[0])
                print('computing properties of converged solution...')
                print('omega = ', omega)
        self.I.ground_state()
        ave = self.I.averages()


        # writes the solution in the standard file
        if root:
            self.I.props['method'] = method
            self.I.props['SEF_eval'] = SEF_eval
            if iH is not None: 
                for i in range(nvar):
                    self.I.props['2der_' + varia[i]] = H[i, i]
            if hartree != None:
                self.I.props['omega'] = omega
            self.I.write_summary(file)
            VCA.first_time = False

        if root:
            pyqcm.banner('VCA ended normally', '*')

        self.hessian = None
        if lib_minimization is False:
            self.hessian = 1.0/np.diag(iH)


    #----------------------------------------------------------------------------------------------------------
    # minimax method

    def _minimax(self, names, var_max_start, start=None, step=None, accur=None, max=10,  max_iteration=30, hartree=None):
        """Applies a minimization of a subset of variables and a maximization on the remainder
        
        :param ['str'] names : names of variational parameters (minimal ones first, then maximal ones)
        :param int var_max_start : label of the first maximal variational parameter (= number of minimal variational parameters)
        :param [float] start: the starting values
        :param [float] step: the steps used to computed the numerical second derivatives
        :param [float] accur: the required accuracy for each variable
        :param [float] max: maximum absolute value of each parameter
        :param int max_iteration: maximum number of iterations, beyond which an exception is raised
        :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
        :returns [float]: the value of the variables at the minimax saddle point

        """

        global current_instance
    
        ftol = 10*pyqcm.qcm.get_global_parameter('accur_SEF')
        nvar_max = len(names) - var_max_start
        nvar_min = var_max_start
        steps_min = step[0:nvar_min]
        steps_max = step[nvar_min:nvar_min+nvar_max]
        iter = 0
        nvar_tot = nvar_min+nvar_max

        def F_min(x):
            global SEF_eval
            for i in range(nvar_min): 
                self.model.set_parameter(names[i], x[i])
            if verbose: print('(min) x = ', x) # new
            SEF_eval += 1
            self.I = pyqcm.model_instance(self.model)
            return self.I.Potthoff_functional(hartree)
        def F_max(x):
            global SEF_eval
            for i in range(nvar_max): 
                self.model.set_parameter(names[i+nvar_min], x[i])
            if verbose: print('(max) x = ', x) # new
            SEF_eval += 1
            self.I = pyqcm.model_instance(self.model)
            return -self.I.Potthoff_functional(hartree)


        X0 = np.array(start)

        while iter < max_iteration:
            P = self.model.parameters()
            x_min = [P[names[i]] for i in range(nvar_min)]
            x_max = [P[names[i]] for i in range(nvar_min, nvar_tot)]

            if iter > 0:
                steps_min = np.abs(diff[0:nvar_min])
                steps_max = np.abs(diff[nvar_min:nvar_min+nvar_max])

            nvar = nvar_min
            steps = steps_min
            initial_simplex = np.zeros((nvar+1,nvar))
            for i in range(nvar+1):
                initial_simplex[i, :] = x_min
            for i in range(nvar):
                initial_simplex[i+1, i] += steps[i]
            if pyqcm.is_sequence(accur) == True:
                accur = accur[0]
            solution = minimize(F_min, x_min, method='Nelder-Mead', options={'maxfev':100, 'xatol': accur, 'fatol':ftol, 'initial_simplex': initial_simplex, 'adaptive': True, 'disp':True})
            iter_done = solution.nit
            sol_min = solution.x
            for i in range(nvar_min):
                self.model.set_parameter(names[i], sol_min[i])
            
            nvar = nvar_max
            steps = steps_max
            initial_simplex = np.zeros((nvar+1,nvar))
            for i in range(nvar+1):
                initial_simplex[i, :] = x_max
            for i in range(nvar):
                initial_simplex[i+1, i] += steps[i]
            if type(accur) == list:
                accur = accur[0]
            solution = minimize(F_max, x_max, method='Nelder-Mead', options={'maxfev':100, 'xatol': accur, 'fatol':ftol, 'initial_simplex': initial_simplex, 'adaptive': True, 'disp':True})
            iter_done = solution.nit
            sol_max = solution.x
            for i in range(nvar_max):
                self.model.set_parameter(names[i+nvar_min], sol_max[i])

            sol = np.concatenate((sol_min, sol_max))

            hartree_converged = True
            if hartree != None:
                hartree_converged = True
                for C in hartree:
                    C.update(current_instance, pr=True)
                    hartree_converged  = hartree_converged and C.converged()

            diff = sol - X0
            X0 = sol
            iter += 1
            converged = True
            for i in range(nvar_tot):
                if np.abs(diff[i]) > accur: converged = False

            if converged and hartree_converged:
                return sol
        
        raise pyqcm.TooManyIterationsError(max_iteration)
        return sol

################################################################################
def plot_sef(model, param, prm, file="sef.tsv", accur_SEF=1e-4, hartree=None, show=True, symmetrized_operator=None, consistency_check=False):
    """Draws a plot of the Potthoff functional as a function of a parameter param taken from the list prm. The results are going to be appended to 'sef.tsv'
    
    :param lattice_model model: the lattice model
    :param str param: name of the parameter (independent variable)
    :param [float] prm: list of values of the parameter
    :param lattice_model model: the lattice model on which the computation is based
    :param float accur_SEF: precision of the computation of the self-energy functional
    :param str file: name of the file to which intermediate results are written
    :param [hartree] hartree: Hartree approximation couplings (see pyqcm/hartree.py)
    :param bool show: if True, the plot is shown on the screen
    :param str symmetrized_operator: name of an operator with respect to which the functional is symmetrized
    :param bool consistency_check: if True, checks the ground state consistency at each point
    :returns: None

    """
    L = model.nsites

    if isinstance(file, str) == False:
	    raise TypeError('the argument "file" of plot_sef() must be a string')
    if isinstance(accur_SEF, float) == False:
	    raise TypeError('the argument "accur_SEF" of plot_sef() must be a float')
    if isinstance(param, str) == False:
	    raise TypeError('the argument "param" of plot_sef() must be a string, the name of a parameter')

    if pyqcm.is_sequence(hartree) == False and hartree is not None:
        hartree = (hartree,)

    pyqcm.set_global_parameter('accur_SEF', accur_SEF)
    omega = np.empty(len(prm))
    for i in range(len(prm)):
        model.set_parameter(param, prm[i])
        I = pyqcm.model_instance(model)
        omega[i] = I.Potthoff_functional(hartree, file=file, symmetrized_operator=symmetrized_operator, consistency_check=consistency_check)
        print("omega(", prm[i], ") = ", omega[i])
    
    if show:
        import matplotlib.pyplot as plt
        plt.xlim(prm[0], prm[-1])
        plt.plot(prm, omega, 'b-')
        plt.xlabel(param)
        plt.ylabel(r'$\Omega$')
        plt.axhline(omega[0], c='r', ls='solid', lw=0.5)
        plt.title(model.parameter_string(), fontsize=6)
        plt.show()




################################################################################
def plot_GS_energy(model, param, prm, clus=0, file=None, plt_ax=None, **kwargs):
    """Draws a plot of the ground state energy as a function of a parameter param taken from the list `prm`. The results are going to be appended to 'GS.tsv'
    
    :param lattice_model model: the lattice model on which the computation is based
    :param str param: name of the parameter (independent variable)
    :param [float] prm: list of values of the parameter
    :param int clus: label of the cluster (starts at 0)
    :param str file: if not None, saves the plot in a file with that name
    :param matplotlib.axes.Axes plt_ax: optional matplotlib axis set, to be passed when one wants to collect a subplot of a larger set
    :param kwargs: keyword arguments passed to the matplotlib 'plot' function
    :returns: None

    """
    import matplotlib.pyplot as plt
    if plt_ax is None:
        plt.figure()
        plt.gcf().set_size_inches(13.5/2.54, 9/2.54)
        ax = plt.gca()
    else:
        ax = plt_ax

    omega = np.empty(len(prm))
    for i in range(len(prm)):
        model.set_parameter(param, prm[i])
        I = pyqcm.model_instance(model)
        omega[i] = I.ground_state()[clus][0]
        print("omega(", prm[i], ") = ", omega[i])

        # writing the parameters in a progress file
        I.write_summary('GS.tsv')

    
    ax.axhline(omega[0], c='r', ls='solid', lw=0.5)
    ax.plot(prm, omega, 'b-')
    ax.set_xlim(prm[0], prm[-1])
    if plt_ax is None:
        ax.set_xlabel(param)
        ax.set_ylabel('GS energy')
        ax.set_title(model.parameter_string())

    if file is not None:
        plt.savefig(file)
        plt.close()
    elif plt_ax is None:
        plt.show()



################################################################################
# detects a continuous phase transition

def _transition(model, varia, P, bracket, step=0.001, verb=False, symmetrized_operator=None):
    r"""Detects a transition as a function of external parameter param by looking at the equality between 
    :math:`\Omega(h=s)` and :math:`\Omega(h=0)` where *h* is a single variational parameter (Weiss field)
    and *s* is a step. 
    
    :param str varia: name of the variational parameter
    :param str P: name of the control parameter (the parameter that controls the transition)
    :param (float,float) bracket: bracketing values for parameter P that enclose the transition
    :param float step: small, but finite value *s* of the Weiss field 
    :param bool verb: If True, prints progress
    :returns: the value of P at the transition

    """

    from scipy.optimize import brentq

    def F(x):
        model.set_parameter(P, x)
        model.set_parameter(varia, step)
        I = pyqcm.model_instance(model)
        Om1 = I.Potthoff_functional(symmetrized_operator=symmetrized_operator)
        model.set_parameter(varia, 1e-8)
        I = pyqcm.model_instance(model)
        Om0 = I.Potthoff_functional(symmetrized_operator=symmetrized_operator)
        if verb:
            print(P, '= ', x, '\tdelta Omega = ', Om1-Om0)
        return Om1-Om0


    x0, r = brentq(F, bracket[0], bracket[1], maxiter=100, full_output=True, disp=True)

    print(r.flag)
    if not r.converged:
        raise RuntimeError('the root finding routine could not find a solution!')
        
    return x0

################################################################################
# detects a continuous phase transition (meta function with loop over control parameter)

def transition_line(model, varia, P1, P1_range, P2, P2_range, delta, verb=False):
    """Builds the second-order transition line as a function of a control parameter P1. The results are written in the file `transition.tsv`

    :param lattice_model model: the lattice model on which the computation is based
    :param str varia: variational parameter
    :param str P1: control parameter
    :param [float] P1_range: an array of values of P1
    :param str P2: dependent parameter
    :param (float) P2_range: 2-uple of values of P2 that bracket the transition
    :param bool verb: If True, prints progress
    :param float delta: at each step, the new bracket for P2 will be P2c +/- delta, P2c being the previous critical value 
    :returns:None

    """

    assert type(varia) == str, 'the first argument of transition_line must be the name of a single variational parameter'
    assert type(P1) == str, 'the second argument of transition_line must be the name of a single control parameter'
    assert type(P2_range) == list, 'the argument P2_range of transition_line must be a list of 2 floats'

    trans = np.empty((len(P1_range), 2))
    for i, p1 in enumerate(P1_range):
        model.set_parameter(P1, p1)
        p2c = _transition(model, varia, P2, P2_range, step=0.001, verb=verb)
        trans[i,0] = p1
        trans[i,1] = p2c
        P2_range[0] = p2c - delta
        P2_range[1] = p2c + delta
        print('---> transition at ', P1, ' = ', p1, '\t', P2, ' = ', p2c)
        np.savetxt('transition.tsv', trans[0:i+1, :], delimiter='\t', header=P1+'\t'+P2)



from pyqcm import *

ns = 2
nb = 4
no = ns+nb
new_cluster_model('L2_4b', ns, nb)

new_cluster_operator('L2_4b', 'tb1', 'one-body', [
    (1, 3, -1.0),
    (2, 4, -1.0),
    (7, 9, -1.0),
    (8, 10, -1.0)
])

new_cluster_operator('L2_4b', 'tb2', 'one-body', [
    (1, 5, -1.0),
    (2, 6, -1.0),
    (7, 11, -1.0),
    (8, 12, -1.0)
])

new_cluster_operator('L2_4b', 'eb1', 'one-body', [
    (3, 3, 1.0),
    (4, 4, 1.0),
    (9, 9, 1.0),
    (10, 10, 1.0)
])

new_cluster_operator('L2_4b', 'eb2', 'one-body', [
    (5, 5, 1.0),
    (6, 6, 1.0),
    (11, 11, 1.0),
    (12, 12, 1.0)
])

new_cluster_operator('L2_4b', 'sb1', 'anomalous', [
    (1, 3+no, -1.0),
    (2, 4+no, -1.0),
    (3, 1+no, 1.0),
    (4, 2+no, 1.0)
])

new_cluster_operator('L2_4b', 'sb2', 'anomalous', [
    (1, 5+no, -1.0),
    (2, 6+no, -1.0),
    (5, 1+no, 1.0),
    (6, 2+no, 1.0)
])

new_cluster_operator_complex('L2_4b', 'sbi1', 'anomalous', [
    (1, 3+no, -1.0j),
    (2, 4+no, -1.0j),
    (3, 1+no, 1.0j),
    (4, 2+no, 1.0j)
])

new_cluster_operator_complex('L2_4b', 'sbi2', 'anomalous', [
    (1, 5+no, -1.0j),
    (2, 6+no, -1.0j),
    (5, 1+no, 1.0j),
    (6, 2+no, 1.0j)
])

new_cluster_operator('L2_4b', 'pb1', 'anomalous', [
    (3, 4+no, 1.0),
    (4, 3+no, -1.0),
    (5, 6+no, 1.0),
    (6, 5+no, -1.0)
])



#-------------------------------------------------------------------
# construction of the lattice model 

add_cluster('L2_4b', [0,0,0], [[0,0,0], [1,0,0]])
lattice_model('1D_2_4b', [[2,0,0]])

interaction_operator('U')
hopping_operator('t', [1,0,0], -1) # NN hopping
hopping_operator('ti', [1,0,0], -1, tau=2) # NN hopping with imaginary amplitude
hopping_operator('tp', [2,0,0], -1) # NNN hopping
hopping_operator('sf', [0,0,0], -1, tau=0, sigma=1) # on-site spin flip
hopping_operator('h', [0,0,0], -1, tau=0, sigma=3) # on-site spin flip
anomalous_operator('D', [1,0,0], 1) # NN singlet
anomalous_operator('Di', [1,0,0], 1j) # NN singlet with imaginary amplitude
anomalous_operator('S', [0,0,0], 1) # on-site singlet 
anomalous_operator('Si', [0,0,0], 1j) # on-site singlet with imaginary amplitude
anomalous_operator('dz', [1,0,0], 1, type='dz') # NN triplet
anomalous_operator('dy', [1,0,0], 1, type='dy') # NN triplet
anomalous_operator('dx', [1,0,0], 1, type='dx') # NN triplet
density_wave('M', 'Z', [1,0,0])
density_wave('pT', 'dz', [1,0,0], amplitude=1, link=[1,0,0])


# from pyqcm.draw_operator import draw_cluster_operator, draw_operator
# draw_cluster_operator('L2_4b', 'pb1')


import pyqcm
from pyqcm.cdmft import CDMFT
import numpy as np

import model_1D_2_4b_2b as M

# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
M.model.set_target_sectors(['N6:S0', 'N6:S0', 'N2:S0'])

# Simulation parameters
M.model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    ty = 0.5
    tb1_1=0.5
    tb2_1=0.4
    eb1_1=1.0
    eb2_1=-1.1
    tb1_2=0.5
    tb2_2=0.4
    eb1_2=1.0
    eb2_2=-1.1
""")

varia = ['eb1_1', 'eb2_1', 'tb1_1', 'tb2_1', 'eb1_2', 'eb2_2', 'tb1_2', 'tb2_2']

X = CDMFT(M.model, varia=varia, wc=[0.5,5,5], grid_type='legendre', accur=1e-3, convergence='self-energy', miniter=1, maxiter=64, depth=1, iteration='fixed_point')


X.I.write_all_hdf5('test.h5')
print('GF written:\n', X.I.cluster_Green_function(0.1j, 1))
I = pyqcm.model_instance(M.model)
I.read_all_hdf5('test.h5')
print('\nGF read:\n', I.cluster_Green_function(0.1j, 1))




"""Variable-parameter self-consistency utilities for pyqcm.

Provides the :class:`variable_parameter` class and helper functions
:func:`variable_parameter_self_consistency` and :func:`variable_parameter_search`
to handle parameters that depend self-consistently on an observable (typically the density).
"""
import numpy as np
import pyqcm

class variable_parameter:
    """This class contains the elements needed to perform computations with parameters that depend on some observable (typically the density). The values of the parameters are adjusted as a function of the observable until some self-consistency is reached.

    attributes:
        - param (str) : name of the parameter that depends on the observable
        - F : function that expresses the parameter as a function of the observable (supplied by the user)
        - obs (str) : name of the observable (a parameter). Default is 'mu'
        - lattice (bool) : True if lattice averages are used
        - diff : difference between successive values of the parameter
        - ave : current value of the observable
        - accur : desired accuracy on the parameter

    """

    size0 = 0

    def __init__(self, param, F, obs='mu', lattice=True, accur=1e-3):
        """

        :param str param: name of the original parameter
        :param F: function that returns the value of the parameter as a function of the observable
        :param str obs: name of the parameter whose average is the observable
        :param bool lattice: if True, the lattice average is used, otherwise the cluster average
        :param float accur: required accuracy of the self-consistent procedure

        """

        self.param = param
        self.value = 0.0
        self.F = F
        self.obs = obs
        self.lattice = lattice
        self.diff = 1e6
        self.ave = 0
        self.accur = accur
        self.iter = 0    

    def update(self, pr=False):
        """Updates the value of the parameter based on the value of the observable
        
        :param bool pr: if True, progress is printed on the screen

        """

        if not self.lattice:
            self.ave = pyqcm.cluster_averages()[self.obs][0]
        else:
            self.ave = pyqcm.averages()[self.obs]
        
        P0 = pyqcm.parameters()[self.param]
        P = self.F(self.ave)
        pyqcm.set_parameter(self.param, P)
        self.diff = P-P0
        self.iter += 1
        if pr:
            print('<{:s}> = {:1.3g},  {:s} --> {:1.3g}, diff = {:1.3g}'.format(self.obs, self.ave, self.param, P, self.diff))


    def compute_obs(self):
        """Computes the observable
        """

        if not self.lattice:
            self.ave = pyqcm.cluster_averages()[self.obs][0]
        else:
            self.ave = pyqcm.averages()[self.obs]
        

    def converged(self):
        """Tests whether the self-consistent procedure has converged

        :return bool: True if the mean-field procedure has converged
        
        """

        if np.abs(self.diff) < self.accur:
            return True
        else:
            return False


    def __str__(self):
        return 'parameter: '+self.param+', observable: <'+self.obs+'>'


################################################################################
def variable_parameter_self_consistency(F, var_params, maxiter=10, file='var_param.tsv'):
    """Performs a task with parameters that depend on an observable (like density) and self-consistently
    finds the correct value of the parameters that is self-consistent with the observable

    :param F: task to perform within the loop
    :param [class variable_parameter] var_params: list of variable_parameters (or single variable_parameters)
    :param int maxiter: maximum number of iterations in the self-consistent procedure
    :param str file: name of the file to which the converged result is written via write_summary()

    """

    global first_time

    if type(var_params) is not list:
        var_params = [var_params]

    pyqcm.banner('Variable parameter self-consistency', c='*', skip=1)
    SC_converged = False
    diff_tot = 1e6

    iter = 0
    while True:
        pyqcm.new_model_instance()
        F()
        iter += 1
        pyqcm.banner('self-consistent iteration {:d}'.format(iter), '-')
        diff_tot = 0
        SC_converged = True
        for i,C in enumerate(var_params):
            C.update(pr=True)
            diff_tot += np.abs(C.diff)
            SC_converged = SC_converged and C.converged()

        print('total difference = {:g}'.format(diff_tot))

        if SC_converged:
            pyqcm.write_summary(file)
            first_time = False
            break

        if iter > maxiter :
            raise RuntimeError('Maximum number of iterations exceeded in self-consistency for variable parameters! Aborting...')

    pyqcm.banner('Variable parameter procedure has converged', c='*')
    

################################################################################
def variable_parameter_search(F, var_param, bracket, file='var_param.tsv'):
    """Performs a root search (Brent algorithm) to find the correct value of the variable parameter
    that matches the observable

    :param F: task to perform within the loop
    :param class variable_parameter var_param: variable_parameter
    :param [float] bracket: list of two values of var_param that bracket the solution
    :param str file: name of the file to which the converged result is written via write_summary()

    """

    from scipy.optimize import brentq

    pyqcm.banner('Variable parameter search (Brent method)', c='*', skip=1)

    def G(x):
        pyqcm.set_parameter(var_param.param, x)
        pyqcm.new_model_instance()
        F()
        var_param.compute_obs()
        return x - var_param.F(var_param.ave)

    x0, r = brentq(G, bracket[0], bracket[1], xtol=var_param.accur, maxiter=100, full_output=True, disp=True)

    print(r.flag)
    if not r.converged:
        raise RuntimeError('the root finding routine could not find a solution!')
    else:
        pyqcm.write_summary(file)
        pyqcm.banner('Variable parameter procedure has converged to {:s} = {:4g}'.format(var_param.param, x0), c='*')
        pyqcm.set_parameter(var_param.param, x0)
        pyqcm.new_model_instance()
    


import pyqcm

class slab:
    """Helper class to define multi-layer systems, typically layers of planes.
    The fundamental unit is a 2D model that is repeated in the z direction
    """
    def  __init__(self, name, nlayer, cluster, sites, superlattice, lattice, thickness = None):
        """Constructor

        :param str name: name of the 2D model
        :param int nlayer: total number of layers in the slab
        :param str cluster: name of the cluster model assembled (limited to one cluster)
        :param list[list[int]] sites: list of sites in each layer
        :param list[list[int]] superlattice: superlattice vectors in 2D
        :param list[list[int]] lattice: lattice vectors in 2D
        :param int thickness: number of inequivalent layers (<= nlayer/2). If None, all layers are different

        """
        self.nlayer = nlayer
        self.name = name
        if thickness is None:
            self.thickness=(nlayer+1)//2
            for i in range(nlayer):
                pyqcm.add_cluster(cluster, [0,0,i], sites)
        else:
            self.thickness=thickness
            if thickness > (nlayer+1)//2:
                print('thickness cannot exceed ', (nlayer+1)//2)
                raise ValueError(f'thickness value cannot exceed {(nlayer+1)//2}')
            if thickness == (nlayer+1)//2 and nlayer%2 :
                l1 = nlayer-thickness+1
            else:
                l1 = nlayer-thickness
            for i in range(thickness):
                pyqcm.add_cluster(cluster, [0,0,i], sites)
            for i in range(thickness, nlayer-thickness):
                pyqcm.add_cluster(cluster, [0,0,i], sites, ref=thickness)
            for i in range(l1, nlayer):
                pyqcm.add_cluster(cluster, [0,0,i], sites, ref=nlayer-i)

        pyqcm.lattice_model('slab_{:s}_{:d}L'.format(name, nlayer), superlattice, lattice)
        self.nband = pyqcm.model_size()[1]//nlayer
    
    def hopping_operator(self, name, link, amplitude, **kwargs):
        """Defines a hopping operator repeated across all layers of the slab.
        Same arguments as the pyqcm.hopping_operator() function.

        :param str name: name of the hopping operator
        :param [int] link: bond vector (3-component); the z-component must be 0 (intra-layer) or 1 (inter-layer)
        :param float amplitude: hopping amplitude multiplier
        :param kwargs: additional keyword arguments passed to pyqcm.hopping_operator() (e.g. orb1, orb2, tau, sigma)

        """
        
        if 'orb1' in kwargs:
            b1 = kwargs['orb1']
        else:
             b1 = 1
        if 'orb2' in kwargs:
            b2 = kwargs['orb2']
        else:
             b2 = 1
        if 'tau' in kwargs:
            tau = kwargs['tau']
        else:
            tau = 1
        if 'sigma' in kwargs:
            sigma = kwargs['sigma']
        else:
            sigma = 0
        
        if link[2] == 0:
            for i in range(self.nlayer):
                pyqcm.hopping_operator(name, link, amplitude, orb1=b1+self.nband*i,  orb2=b2+self.nband*i, tau=tau, sigma=sigma)
        elif link[2] == 1:
            for i in range(self.nlayer-1):
                pyqcm.hopping_operator(name, link, amplitude, orb1=b1+self.nband*i,  orb2=b2+self.nband*(i+1), tau=tau, sigma=sigma)
        else:
            raise ValueError('hopping operators in slabs can only have a z-component of 0 or 1 for the link')
        
    def interaction_operator(self, name, **kwargs):
        """Defines an interaction operator repeated across all layers of the slab.
        Same arguments as the pyqcm.interaction_operator() function.

        :param str name: name of the interaction operator
        :param kwargs: additional keyword arguments passed to pyqcm.interaction_operator()
            (e.g. orb1, orb2, type, link, amplitude)

        """
        
        if 'orb1' in kwargs:
            b1 = kwargs['orb1']
        else:
             b1 = 1
        if 'orb2' in kwargs:
            b2 = kwargs['orb2']
        else:
             b2 = 1
        if 'type' in kwargs:
            the_type = kwargs['type']
        else:
            the_type = 'Hubbard'
        if 'link' in kwargs:
            link = kwargs['link']
        else:
            link = ( 0, 0, 0)
        if 'amplitude' in kwargs:
            amplitude = kwargs['amplitude']
        else:
            amplitude = 1.0
       
        if link[2] == 0:
            for i in range(self.nlayer):
                pyqcm.interaction_operator(name, link=link, amplitude=amplitude, orb1=b1+self.nband*i,  orb2=b2+self.nband*i, type=the_type)
        elif link[2] == 1:
            for i in range(self.nlayer-1):
                pyqcm.interaction_operator(name, link=link, amplitude=amplitude, orb1=b1+self.nband*i,  orb2=b2+self.nband*i, type=the_type)
        else:
            raise ValueError('hopping operators in slabs can only have a z-component of 0 or 1 for the link')




import pyqcm
import pyqcm.vca as vca
import numpy as np

import matplotlib.pyplot as plt

# A simple 1D 4 site cluster
CM = pyqcm.cluster_model( 4, 0)
clus = pyqcm.cluster(CM, ((0,0,0),(1,0,0),(2,0,0),(3,0,0)))
model = pyqcm.lattice_model('1D_4', clus, ((4,0,0),))

model.interaction_operator('U', amplitude=1) # on site interaction
model.interaction_operator('V', link=(1,0,0), amplitude=1) # NN interaction
model.hopping_operator('t', [1,0,0], -1) 
model.density_wave('Delta', 'N', (1,0,0), amplitude=1) # A charge density wave with period 2 in position space

# Defining an explicit operator for the "Hartree part" of the extended interaction
elems = [((0,0,0), (0,0,0), 1/np.sqrt(2)), ((3,0,0), (0,0,0), 1/np.sqrt(2))] # the 1/sqrt(2) factor ensures normalization
model.explicit_operator('Vm1', elems, type='one-body', tau=0) # Vm is an on-site operator ---> tau=0

elems = [((0,0,0), (0,0,0), 1/np.sqrt(2)), ((3,0,0), (0,0,0), -1/np.sqrt(2))] # the 1/sqrt(2) factor ensures normalization
model.explicit_operator('Vm2', elems, type='one-body', tau=0) # Vm is an on-site operator ---> tau=0

# Parameters required for half-filling
model.set_target_sectors('R0:N4:S0')
model.set_parameters("""
    U=4
    V=4
    Vm1=5
    mu=10
    t=1
    Delta_1=1e-9
    Delta=1e-9
""")
I = pyqcm.model_instance(model)

# Defining an object of the hartree class with eigenvalue 1
MF1 = pyqcm.hartree(model, 'Vm1', 'V', 1, accur=1e-2, lattice=True) # lattice=True ---> use lattice averages

#########################################################################################
############################## - Range for sweep over V - ###############################

V_start = 1e-9
V_stop = 4
V_step = 0.1
U = model.parameters()['U']

#########################################################################################
############################# - Self-consistency approach - #############################

# This is simply a loop that applies the self_consistency approach over V_range

def F():
    model.set_parameter('V', V)
    model.set_parameter('mu', 2*V + U/2) # Condition to impose half-filling
    return pyqcm.model_instance(model)

# for V in np.arange(V_start, V_stop, V_step):
#     model.Hartree_procedure(F, MF1, maxiter=32)

#########################################################################################
################################### - VCA approach - ####################################

# This is the function to run inside of controlled_loop to perform the vca itself
def run_vca():
    V = model.parameters()['V']
    model.set_parameter('mu', 2*V + U/2) # Condition to impose half-filling
    V = vca.VCA(model, varia='Vm1', steps=0.05, accur=2e-3, max=20, hartree=MF1, max_iter=64)    
    return V.I

# performing the vca loop over V_range
model.controlled_loop(run_vca, varia='Vm1', loop_param='V', loop_range=(V_start, V_stop, V_step))

#########################################################################################
############################# - VCA over both parameters - ##############################

# Function to perform the VCA
varia=('Vm1', 'Delta_1')
def run_vca_2():
    V = model.parameters()['V']
    model.set_parameter('mu', 2*V + U/2) # Condition to impose half-filling
    V = vca.VCA(model, varia=varia, steps=0.05, accur=2e-3, max=20, hartree=MF1, max_iter=64)    
    return V.I

# Looping from 4 to 3
model.controlled_loop(run_vca_2, varia=varia, loop_param='V', loop_range=(V_start, V_stop, V_step))

#########################################################################################
######################### - Hybrid VCA with self-consistency - ##########################

def run_vca_3():
    V = model.parameters()['V']
    model.set_parameter('mu', 2*V + U/2) # Condition to impose half-filling
    V = vca.VCA(model, varia='Delta_1', steps=0.05, accur=2e-3, max=20, max_iter=64, hartree= MF1, hartree_self_consistent=True)    
    return V.I

model.controlled_loop(run_vca_3, varia=('Delta_1', 'Vm1'), loop_param='V', loop_range=(V_start, V_stop, V_step))




import pyqcm
from pyqcm.vca import *
import numpy as np

from model_hartree import model

# Parameters required for half-filling
model.set_target_sectors(["R0:N4:S0"])
model.set_parameters("""
    U=4
    V=4
    Vm=5
    mu=10
    t=1
    Delta_1=1e-9
    Delta=1e-9
""")


# Updates mu as a function of V and U
def _adjust_mu():
    V = model.parameters()['V']
    U = model.parameters()['U']
    model.set_parameter('mu', 2*V + U/2) # Condition to impose half-filling
    
# Defining an object of the hartree class with eigenvalue 1
Vm_H = pyqcm.hartree(model, 'Vm', 'V', 1, lattice=True) # lattice=True ---> use lattice averages

# Range for sweep over V 
V_start = 4
V_stop = 2
V_step = -0.5

def F():
    return pyqcm.model_instance(model)

############################# - Self-consistency approach - #############################
alpha = 0.0

# This is simply a loop that applies the self_consistency approach over V_range
for V in np.arange(V_start, V_stop, V_step):
    model.set_parameter('V', V)
    _adjust_mu()

    model.Hartree_procedure(F, Vm_H, iteration='fixed_point')

for V in np.arange(V_start, V_stop, V_step):
    model.set_parameter('V', V)
    _adjust_mu()

    I, niter, alpha = model.Hartree_procedure(F, Vm_H, iteration='Broyden', alpha = alpha)
################################### - VCA approach - ####################################

# This is the function to run inside of controlled_loop to perform the vca itself
def _run_vca():
    _adjust_mu()
    V = VCA(model, varia=['Vm'], steps=[0.05], accur=[2e-3], max=[20], hartree=Vm_H, max_iter=300)
    return V.I 


model.set_parameter('Vm', 5.65) # Initial guess for the VCA based on the Hartee self-consistency
_adjust_mu()

# performing the vca loop over V_range
model.controlled_loop(_run_vca, varia='Vm', loop_param='V', loop_range=(V_start, V_stop, V_step))



import pyqcm
from pyqcm.cdmft import *

def model1D(ns):
    g = [ns-i for i in range(ns)]
    CM = pyqcm.cluster_model(ns, generators=[g])  
    clus = pyqcm.cluster(CM, [(i,0,0) for i in range(ns)]) 
    return pyqcm.lattice_model('1D_{:d}'.format(ns), clus, ((ns,0,0),)) 

urange = np.arange(1e-9, 16.1, 2)
lrange = np.array([2,4,6,8,10], dtype=int)

f = np.empty((len(lrange), len(urange)))

with open('fid.tsv', 'w') as fout:
    for i,L in enumerate(lrange):
        pyqcm.reset_model()
        model = model1D(int(L))
        model.interaction_operator('U')
        model.hopping_operator('t', (1,0,0), -1)

        model.set_target_sectors(['R0:N{:d}:S0'.format(model.nsites)])
        model.set_parameters("""
            U = 1e-9
            mu = 0.5*U
            t = 1
        """)

        I1 = pyqcm.model_instance(model) 
        for j,u in enumerate(urange):
            model.set_parameter('U', u)
            I2 = pyqcm.model_instance(model) # call to constructor of class "model_instance"
            f[i,j] = model.fidelity(I1,I2)
            S = '{:d}\t{:1.2f}\t{:g}'.format(L, u, f[i,j])
            print(S)
            fout.writelines(S+'\n')
        fout.writelines('\n\n')
        

    for j,u in enumerate(urange):
        pol = np.polyfit(1/lrange, f[:,j], 2)
        fex = np.polyval(pol, 0.0)
        S = 'extr\t{:1.2f}\t{:g}'.format(u, fex)
        print(S)
        fout.writelines(S+'\n')
    



import pyqcm
import pyqcm.vca as vca
import numpy as np


pyqcm.set_global_parameter("verb_integrals")

# declare a cluster model of 9 sites, named 'clus'
CM = pyqcm.cluster_model(9)

sites = []
for j in range(3):
    for i in range(3):
        sites.append((i,j,0))

# defining a hopping operator that only hops around the perimeter of the clusters
ns = 9
CM.new_operator('tperim', 'one-body',[
    (1, 2, -1.0), 
    (2, 3, -1.0),
    (3, 6, -1.0), 
    (6, 9, -1.0),
    (8, 9, -1.0), 
    (7, 8, -1.0),
    (4, 7, -1.0), 
    (1, 4, -1.0),
    (1+ns, 2+ns, -1.0), 
    (2+ns, 3+ns, -1.0),
    (3+ns, 6+ns, -1.0), 
    (6+ns, 9+ns, -1.0),
    (8+ns, 9+ns, -1.0), 
    (7+ns, 8+ns, -1.0),
    (4+ns, 7+ns, -1.0), 
    (1+ns, 4+ns, -1.0)
])

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus1 = pyqcm.cluster(CM, sites) 
clus2 = pyqcm.cluster(CM, sites, (3,0,0)) # second cluster, offset from the first

# define a lattice model named '1D_4' made of the cluster(s) clus and superlattice vector (4,0,0)
model = pyqcm.lattice_model('model_3x3_2C', (clus1, clus2), ((6,0,0),(1,3,0)))

# define a few operators in this model
model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)
model.hopping_operator('t', (0,1,0), -1)
model.density_wave('M', 'Z', (1,1,0)) # Spin density wave at Q = (pi,pi)

model.set_target_sectors(('R0:N9:S-1', 'R0:N9:S1')) # half filling for both clusters
model.set_parameters("""
t = 1
U = 8
mu = 0.5*U
M = 0
M_1 = 0.1
M_2 = 1*M_1
t_1 = 1
t_2 = 1*t_1
tperim_1 = 1e-9
tperim_2 = 1*tperim_1
""")

V = vca.VCA(model, varia=('M_1', 't_1', 'tperim_1'), start=(0.17, 1.15, 0.04), steps=(0.005, 0.005, 0.005), accur=(2e-3, 2e-3, 2e-3), max=(10, 10, 10))




import pyqcm
import model_1D_4_C2 as M
import numpy as np

M.model.set_target_sectors(['R0:N4:S0'])
M.model.set_parameters("""
t=1
U = 4
mu = 8
""")

##################################################################
# TEST UNITAIRE


def test_fixed_density_loop():

    pyqcm.banner('testing fixed_density_loop()', c='#', skip=1)

    def F():
        return pyqcm.model_instance(M.model)

    M.model.fixed_density_loop(
        F,
        1.1,  # target density
        5,  # starting  value of mu
        kappa=1.0,
        maxdmu=0.5,  # maximum change in mu
        loop_param='U',
        loop_values=np.arange(6, 4, -0.2),
        dens_tol=0.002,
        dir='',
        measure=None
    )


# test_fixed_density_loop()



import pyqcm
import numpy as np

# Constructing a new cluster model
ns = 2 # number of physical sites
nb = 4 # number of bath sites
no = ns+nb # total number of sites
CM = pyqcm.cluster_model( ns, nb)

########################### BATH OPERATORS ########################### 
# Defining the bath hopping operators
CM.new_operator('tb1', 'one-body', [
    (1, 3, -1.0),
    (2, 4, -1.0),
    (1+no, 3+no, -1.0),
    (2+no, 4+no, -1.0)
]) # note that the last two entries pertain to the SPIN DOWN part of the operator

CM.new_operator('tb2', 'one-body', [
    (1, 5, -1.0),
    (2, 6, -1.0),
    (1+no, 5+no, -1.0),
    (2+no, 6+no, -1.0)
])

# Defining the 'orbital energy' of the baths
CM.new_operator('eb1', 'one-body', [
    (3, 3, 1.0),
    (4, 4, 1.0),
    (3+no, 3+no, 1.0),
    (4+no, 4+no, 1.0)
])

CM.new_operator('eb2', 'one-body', [
    (5, 5, 1.0),
    (6, 6, 1.0),
    (5+no, 5+no, 1.0),
    (6+no, 6+no, 1.0)
])

######################################################################
clus = pyqcm.cluster(CM, ((0, 0, 0), (1, 0, 0)))
model = pyqcm.lattice_model('Graphene_2', clus, ((1,-1, 0), (2, 1, 0)), ((1,-1, 0), (2, 1, 0)))
model.set_basis([( 1, 0, 0),[-0.5,np.sqrt(3)/2,0]]) # Classic Graphene basis (for simplicity and graphical purposes)

# Defining the interaction operator on BOTH bands
model.interaction_operator('U', orbitals=(1,1))
model.interaction_operator('U', orbitals=(2,2))

# Defining NN hopping terms
model.hopping_operator('t', (1,0,0), -1, orbitals=(1,2)) # All hops here are from one band to another
model.hopping_operator('t', (0,1,0), -1, orbitals=(1,2))
model.hopping_operator('t', (-1,-1,0), -1, orbitals=(1,2))

model.hopping_operator('M', (0,0,0),  1, orbitals=(1,1), tau=0, sigma=3)
model.hopping_operator('M', (0,0,0), -1, orbitals=(2,2), tau=0, sigma=3)
model.hopping_operator('Mx', (0,0,0),  1, orbitals=(1,1), tau=0, sigma=1)
model.hopping_operator('Mx', (0,0,0), -1, orbitals=(2,2), tau=0, sigma=1)
model.hopping_operator('cdw', (0,0,0),  1, orbitals=(1,1))
model.hopping_operator('cdw', (0,0,0), -1, orbitals=(2,2))



import numpy as np
import pyqcm
from pyqcm.cdmft import CDMFT

from model_G4_6b_2C import model

# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
model.set_target_sectors(['R0:N10:S0']*2)

# Simulation parameters
model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    t_1 = 1.1
    t_2 = 0.9
    tb1_1=0.5
    tb2_1=1*tb1_1
    eb1_1=1.0
    eb2_1=-1.0*eb1_1
    tb1_2= -1*tb1_1
    tb2_2= -1*tb2_1
    eb1_2=1*eb1_1
    eb2_2=1*eb2_1
""")

I = pyqcm.model_instance(model)  
print('\nstrict hopping matrix of cluster 1 :')              
A = I.cluster_hopping_matrix(clus=0)
print(np.real(A))

print('\nstrict hopping matrix of cluster 2 :')              
A = I.cluster_hopping_matrix(clus=1)
print(np.real(A))

print('\ninteractions of cluster 1 :')              
print(I.interactions(sys=0))
print('\ninteractions of cluster 2 :')              
print(I.interactions(sys=1))


I.write_impurity_problem(0, bath_diag=False, file='impurity1.tsv')
I.write_impurity_problem(1, bath_diag=False, file='impurity2.tsv')

print('\ncluster Green function of cluster 1 at z=0.1j:')
print(I.cluster_Green_function(0.1j, clus=0, spin_down=False, blocks=False))
print('\ncluster Green function of cluster 2 at z=0.1j:')
print(I.cluster_Green_function(0.1j, clus=1, spin_down=False, blocks=False))

print('\ncluster self-energy of cluster 1 at z=0.1j:')
print(I.cluster_self_energy(0.1j, clus=0, spin_down=False))
print('\ncluster self-energy of cluster 2 at z=0.1j:')
print(I.cluster_self_energy(0.1j, clus=1, spin_down=False))

print('\nGreen function average of cluster 1:')
print(I.Green_function_average(sys=0, spin_down=False))
print('\nGreen function average of cluster 2:')
print(I.Green_function_average(sys=1, spin_down=False))




import matplotlib.pyplot as plt
import numpy as np
import sys

if __name__ == "__main__":
  
  f_integrals_small = "small/integrals_small_results.txt"
  data = np.genfromtxt(f_integrals_small, delimiter=",", comments='#',skip_header=4)
  
  fig,ax = plt.subplots()
  colors = ['r','b','g']
  
  labels = ["MKL","OPENBLAS","BLIS"]
  
  for j in range(1,data.shape[1]):
      ax.plot(data[:,0], data[:,j], color=colors[j-1], label=labels[j-1], marker='o')
  print(data)
  max_core = int(data[0,0])
  ref_time = data[-1,1]
  ax.plot([x+1 for x in range(max_core)],[ref_time/(x+1) for x in range(max_core)], linestyle='--', color='k', label='Ideal')
  
  ax.set_xlabel("Cores")
  ax.set_ylabel("Time (s)")
  ax.grid()
  ax.set_ylim(4,300)
  ax.set_xlim([1,96])
  ax.set_yscale("log", base=2)
  ax.set_xscale("log", base=2)
  ax.set_yticks([int(2**x) for x in range(2,9)])
  ax.set_yticklabels([str(int(x)) for x in ax.get_yticks()])
  ax.set_xticks([int(2**x) for x in range(0,8)])
  ax.set_xticklabels([str(int(x)) for x in ax.get_xticks()])
  ax.legend()
  
  plt.tight_layout()
  plt.show()




####################################################################################################
# Simple Hubbard model on a periodic cluster
# We Define a simple nearest-neighbor, half-filled Hubbard model with a periodic cluster of length 12 in dimension 1. Make use of translation symmetry. For each value of the 12 possible values of crystal momentum in this system, compute the ground state energy of the cluster for $U=0$ and $U=4$. This problem concerns the cluster only and makes no use of CPT or other quantum cluster methods.
####################################################################################################
import pyqcm
import numpy as np

import matplotlib.pyplot as plt
import matplotlib as mpl

# global option that toggles periodic boundary conditions
pyqcm.set_global_parameter("periodic") 

# defines the cluster with a translation symmetry generator 
CM = pyqcm.cluster_model(12, generators = ((12, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11),)) 

# define a physical cluster based on that model, with base position (0,0,0) and site positions
clus = pyqcm.cluster(CM, [(i,0,0) for i in range(12)]) 

# define a lattice model 
model = pyqcm.lattice_model('periodic12', clus, ((12,0,0),))

# define a few operators in this model. Since this is a 2-orbital model, one needs to specify the relevant orbitals.
model.interaction_operator('U')
model.hopping_operator('t', (1,0,0), -1)

model.set_parameters("""
    t=1
    U=4
    mu=0.5*U
""")

# loop over the different irreducible representations (number of electrons and wavevectors)

for n in range(1,13):
    for k in range(12):
        model.set_target_sectors([f"R{k}:N{n}:S{n%2}"]) 
        I = pyqcm.model_instance(model)
        print(f"N={n}\tk={k}\tE0={I.ground_state(file='GS.tsv')[0][0]}")



import pyqcm
import numpy as np
from model_1D import model1D
model = model1D(4, sym=False)

pyqcm.set_global_parameter('accur_OP', 1e-8)
sec = 'R0:S0'

print(model.currents)

phase = np.pi/5
model.set_target_sectors([sec])
model.set_parameters("""
t={:g}
ti={:g}
U = 4
mu = 1
D = 0.5
It = 0
Iti = 0
""".format(np.cos(phase), np.sin(phase)))

P = model.parameters()
print(P)

I = pyqcm.model_instance(model)
# ave = I.averages(pr=True)
# print('current = ', ave['It']*P['t'] + ave['Iti']*P['ti'] )




import pyqcm
from pyqcm.vca import *
# pyqcm.set_global_parameter("nosym")
import model_1D_4



class subsystem:
    def __init__(self, basis, A, B):
        self.basis = basis
        self.A = A
        self.B = B
        self.L = A.shape[0]+B.shape[0]  # total number of sites
        self.maskA = 0
        for i in range(A.shape[0]):
            self.maskA += 1 << A[i]
            self.maskA += 1 << (A[i]+self.L)
        self.maskB = 0
        for i in range(B.shape[0]):
            self.maskB += 1 << B[i]
            self.maskB += 1 << (B[i]+self.L)
        self.maskA = np.uint(self.maskA)
        self.maskB = np.uint(self.maskB)

    def transpose(self, I, Ip):
        As = self.basis[I] & self.maskA
        Bs = self.basis[I] & self.maskB
        Asp = self.basis[Ip] & self.maskA
        Bsp = self.basis[Ip] & self.maskB
        K = np.where(self.basis == As + Bsp)
        Kp = np.where(self.basis == Asp + Bs)
        return K[0][0], Kp[0][0]

    def transpose_matrix(self, A):
        At = np.empty(A.shape)
        for I in range(basis.shape[0]):
            for J in range(I+1):
                Ip, Jp = S.transpose(I,J)
                At[Ip, Jp] = A[I,J]
                At[Jp, Ip] = A[J,I]
        return At





pyqcm.set_global_parameter("verb_ED")

pyqcm.set_target_sectors(['R0:S0'])
pyqcm.set_parameters("""
t=1
U=4
mu = 2
D = 0.4
""")
pyqcm.new_model_instance()
sites = [0,1]
rho, basis = pyqcm.density_matrix(sites)
print(np.real(rho))
print('trace of rho : ', np.trace(rho))

L = len(sites)
print("basis of subsystem:\n")
for i in range(basis.shape[0]):
    print('|' + np.binary_repr(basis[i], 2*L)+'>')


S = subsystem(basis, np.array([0]), np.array([1]))

# for I in range(basis.shape[0]):
#     for J in range(basis.shape[0]):
#         Ip, Jp = S.transpose(I,J)
#         print('({:d}, {:d}) --> ({:d}, {:d})'.format(I,J,Ip,Jp))

rhoT = S.transpose_matrix(rho)

print('spectrum of rhoT:\n', np.linalg.eigh(rhoT)[0])

        


    


####################################################################################################
# Antiferromagnetism with various cluster sizes
# We construct several clusters to study the simple Hubbard model in two dimension, with sizes 2, 4, 6, 8 and 10, with the most symmetric shape possible and appropriate superlattices. Arrange the clusters in a way that a study of Néel antiferromagnetism is possible, i.e., respect the bipartite character of the square lattice in the tiling.
# For each of these models, perform a VCA analysis of antiferromagnetism at half-filling for $U=8$ and $t=1$. How does the Weiss field change according to cluster size?
####################################################################################################
import pyqcm
from pyqcm.vca import VCA

def create_model(L):
    """
    param L : length of the cluster (L x 2)
    """

    CM = pyqcm.cluster_model(2*L, name='clus'.format(L))
    sites = [(i,0,0) for i in range(L)] + [(i,1,0) for i in range(L)]
    clus = pyqcm.cluster(CM, sites)
    model = pyqcm.lattice_model('{:d}x2_AFM'.format(L), clus, ((L, L%2, 0), (0, 2, 0)))
    model.hopping_operator('t', [1,0,0], -1)
    model.hopping_operator('t', [0,1,0], -1)
    model.interaction_operator('U')
    model.density_wave('M', 'Z', [1,1,0])
    model.set_target_sectors(['R0:N{:d}:S0'.format(2*L)])
    return model

for L in [1,2,3,4,5]:
    pyqcm.reset_model()
    model = create_model(L)

    # Setting model parameters
    model.set_parameters("""
        U = 8
        mu = 0.5*U
        t = 1
        M = 0
        M_1 = 0.1
    """)

    # Running the VCA
    V = VCA(model, varia='M_1', steps=0.01, accur=2e-4, max=10.0, method='altNR', file = 'vca_L{:d}.tsv'.format(L)) 



import pyqcm
import matplotlib.pyplot as plt
import model_WSM as M
import numpy as np
I = pyqcm.model_instance(M.model)

# ax = plt.gca()
ax = None
F = None

#-----------------------------------------------------------------
# import record_WSM
# k = np.array([0.695,0,-1e-8])
# Berry_curvature(k_perp=0.01, range=[0.5, 0, 0.5], plane='xz', file=F, plt_ax=ax)
# Berry_field_map(k_perp=0.01, plane='xy', file=F, plt_ax=ax)
# Berry_flux(k, 0.1)
# print('Berry flux : ', Berry_flux(k, 0.1))
# Berry_flux_map(nk=100, k_perp=-0.01, file=F, plt_ax=ax)
# print('Chern number : ', Chern_number())
# print('monopole at ', k, ' : ', monopole(k, a=0.05, nk=40))
# monopole_map(plane='xy', k_perp=0.0, file=F, plt_ax=ax)

# plt.show()
##################################################################
# TEST UNITAIRE

def test_berry():
    x = None
    # import record_WSM
    k = np.array([0.695,0,-1e-8])

    F = 'test_Berry_curvature.pdf'
    pyqcm.banner('testing Berry_curvature()', c='#', skip=1); 
    I.Berry_curvature(nk=20, k_perp=0.01, range=[0.5, 0, 0.5], plane='xz', file=F, plt_ax=ax)

    F = 'test_Berry_field_map.pdf'
    pyqcm.banner('testing Berry_field_map()', c='#', skip=1); 
    I.Berry_field_map(nk=20, k_perp=0.01, plane='xy', file=F, plt_ax=ax)

    pyqcm.banner('testing Berry_flux()', c='#', skip=1); 
    print('Berry flux : ', I.Berry_flux(k, 0.1))
    
    F = 'test_Berry_flux_map.pdf'
    pyqcm.banner('testing Berry_flux_map()', c='#', skip=1); 
    I.Berry_flux_map(nk=20, k_perp=-0.01, file=F, plt_ax=ax)
    
    pyqcm.banner('testing Chern_number()', c='#', skip=1); 
    print('Chern number : ', I.Chern_number())
    
    if I.model.dim == 3:
        pyqcm.banner('testing monopole()', c='#', skip=1); 
        print('monopole at ', k, ' : ', I.monopole(k, a=0.05, nk=20))
    
        F = 'test_monopole_map.pdf'
        pyqcm.banner('testing monopole_map()', c='#', skip=1); 
        I.monopole_map(nk=20, plane='xy', k_perp=0.0, file=F, plt_ax=ax)


test_berry()


import pyqcm
from pyqcm.spectral import *
import model_1D_4

# pyqcm.set_global_parameter('verb_ED')

pyqcm.set_target_sectors(['R0:N4:S0'])
pyqcm.set_parameters("""
t=1
U=4
mu=2
""")


pyqcm.new_model_instance()

spectral_function(path='line')


import pyqcm
from model_1D_2_4b import model

# L=6
# model = model1D(L)
# model.set_target_sectors('R0:N{:d}:S0'.format(L))
model.set_target_sectors('R0:N6:S0')
model.set_parameters("""
    U=4
    mu=0.5*U
    t=1
    eb1_1 = 1
    eb2_1 = -1
    tb1_1 = 0.5
    tb2_1 = 0.5
""")

I = pyqcm.model_instance(model)
print(I.cluster_Green_function(0.1j, 0)[0,0])
I.cluster_averages(pr=True)
# I.write_impurity_problem(clus=0, file='impurity.tsv')
# I.write_Green_function(clus=0, file='GF.tsv')
I.write_hdf5('test.h5')
I = pyqcm.model_instance(model)
I.read_hdf5('test.h5')
print(I.cluster_Green_function(0.1j, 0)[0,0])
I.cluster_averages(pr=True)



import pyqcm
import model_1D_2_4b as M

sec = 'R0:N4:S0'
M.model.set_target_sectors([sec])
M.model.set_parameters("""
t=1
U = 4
mu = 0.5*U
eb1_1 = 0.5
eb2_1 = -0.5
tb1_1 = 0.5
tb2_1 = 0.5
""")

I = pyqcm.model_instance(M.model)
I.cluster_spectral_function(wmax=6, opt='hyb', file="cluster_spectral_hybridization.pdf")

# print(pyqcm.hybridization_Lehmann())




import pyqcm
import model_1D_4_C2 as M
import matplotlib.pyplot as plt

mixing = 0

if mixing == 4:
    M.model.set_target_sectors(['R0:N4:S0'])
    M.model.set_parameters("""
    U=1
    t=1
    H = -1
    mu = 1
    """)

elif mixing == 3:
    M.model.set_target_sectors(['R0'])
    M.model.set_parameters("""
    U=1
    t=1
    S = 0.4
    H = 0.5
    Hx = 1
    mu = 1
    """)

elif mixing == 2:
    M.model.set_target_sectors(['R0:N4'])
    M.model.set_parameters("""
    U=1
    t=1
    H = -1
    Hx = 1e-8 
    mu = 1
    """)

elif mixing == 1:
    M.model.set_target_sectors(['R0:S0'])
    M.model.set_parameters("""
    U=1
    t=1
    S = 0.4
    H = -1
    mu = 1
    """)

elif mixing == 0:
    M.model.set_target_sectors(['R0:N4:S0'])
    M.model.set_parameters("""
    U=1
    t=1
    mu = 1
    """)

I = pyqcm.model_instance(M.model)
I.cluster_spectral_function(wmax = 4, plt_ax=plt.gca(), file="test_mixing_nocf.pdf")
# pyqcm.set_global_parameter('continued_fraction')
M.model.set_parameter('U', 1)
I = pyqcm.model_instance(M.model)
I.cluster_spectral_function(wmax = 4, plt_ax=plt.gca(), file="test_mixing_cf.pdf", color='r')




import pyqcm
from pyqcm.spectral import *
# pyqcm.set_global_parameter("nosym")
import model_1D_4

pyqcm.set_global_parameter("continued_fraction")

pyqcm.set_target_sectors(['R0:N4:S0'])
pyqcm.set_parameters("""
t=1
U=2
mu = 1
""")

DoS(w=5, plt_ax=0)
# spectral_function(path="line")




import pyqcm
from model_1D import model1D
model = model1D(4, sym=False)

sec = 'R0:S0'
model.set_target_sectors([sec])
model.set_parameters("""
t=1
U = 4
mu = 1
D = 0.2
""")
I = pyqcm.model_instance(model)

print("\ntesting lattice averages (all operators, also with Potthoff functional and potential energy):\n")
print('Potthoff functional = ', I.Potthoff_functional())
print('potential energy = ', I.potential_energy())
I.averages(pr=True)

print("\ntesting cluster averages from a grid:\n")
wr, weight = pyqcm.legendre_frequency_grid(1,10,10)
# wr, weight = pyqcm.regular_frequency_grid(20, 200, 10)
pyqcm.qcm.frequency_grid(wr, weight)
pyqcm.set_global_parameter('kgrid_side', 128)
I = pyqcm.model_instance(model)
I.averages(pr=True)

I.cluster_averages(pr=True)

print("\ntesting cluster averages from the wavefunction:\n")
I.cluster_averages(pr=True)

print("\ntesting cluster Green function averages:\n")
ave = I.Green_function_average()
print(r'\naverages of c^\dagger_i c_j :\n\n', ave)

if model.mixing == 4:
    ave = I.Green_function_average(spin_down=True)
    print(r'\naverages of c^\dagger_i c_j (spin down):\n\n', ave)

print('\naverage of t from GF= ', -(ave[0,1]+ave[1,2]+ave[2,3]).real)
print('average of mu from GF = ', I.Green_function_density())

I.GS_consistency()



import pyqcm
import pyqcm.vca as V
import matplotlib.pyplot as plt
import numpy as np

# ax = plt.gca()
ax = None
F = None

# print(dir(pyqcm.vca)); exit()
#-----------------------------------------------------------------

# plt.show()
##################################################################
# TEST UNITAIRE

def test_vca():
    x = None
    import model_2x2_C2_vca as M

    F = 'test_vca.pdf'

    pyqcm.banner('testing vca()', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 't_1'], accur=[5e-4, 5e-4], steps=[5e-5, 5e-5], max=[10,10], method='NR')

    pyqcm.banner('testing vca() with Nelder-Mead', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 't_1'], accur=(5e-4, 5e-4), steps=(5e-5, 5e-5), max=(10,10), method='Nelder-Mead')

    pyqcm.banner('testing vca() with Newton-Raphson', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 't_1'], accur=[5e-4, 5e-4], steps=[5e-5, 5e-5], max=[10,10], method='NR')

    pyqcm.banner('testing vca() with explicit starting values', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 't_1'], start=[0.1, 1.3], accur=[5e-4, 5e-4], steps=[5e-5, 5e-5], max=[10,10], method='SYMR1')

    pyqcm.banner('testing vca() with minimax', c='#', skip=1)
    vca = V.VCA(M.model, varia=['M_1', 'mu_1'], accur=[5e-4, 5e-4], steps=[5e-5, 5e-5], max=[10,10], method='minimax', var_max_start=1)

    pyqcm.banner('testing plot_sef()', c='#', skip=1)
    V.plot_sef(M.model, 'M_1', np.arange(1e-9, 0.3, 0.02), accur_SEF=1e-4, show=False)

    pyqcm.banner('testing plot_GS_energy()', c='#', skip=1)
    V.plot_GS_energy(M.model, 'M_1', np.arange(1e-9, 0.3, 0.02), file="GS_energy.pdf")

    # pyqcm.banner('testing transition_line()', c='#', skip=1) ########## Temporary fix ##########
    # M.model.set_parameter('U', 1)
    # M.model.set_parameter('t2', -0.5)
    # V.transition_line(M.model, 'M_1', 'U', np.arange(1, 8.01, 0.25), 'mu', [0.8, 1], delta=1, verb=True)


test_vca()

# transition
# transition_line
# vca_min


import pyqcm
from pyqcm.cdmft import CDMFT

import model_graphene_bath as M

# Imposing half-filling at 6 particles in cluster + bath sites and setting total spin to 0
M.model.set_target_sectors(['R0:N6:S0'])

# Simulation parameters
M.model.set_parameters("""
    U=4
    mu=2
    t=1
    tb1_1=0.5
    tb2_1=0.4
    eb1_1=1.0
    eb2_1=-1.1
""")

varia = ['eb1_1', 'eb2_1', 'tb1_1', 'tb2_1']

convergence='self-energy'; accur = 1e-4

# Defining a function that will run a cdmft procedure within controlled_loop()
def run_cdmft(method):
    try:
        alpha = X.alpha
    except:
        alpha = 0.0
    X = CDMFT(M.model, varia=varia, wc=10, selfnorm=True, accur=accur, convergence=convergence, converge_with_stdev=False, miniter=1, maxiter=64, depth=1, iteration='Broyden', alpha=alpha, method=method)
    return X.I

# Looping over methods

methods = ['Nelder-Mead', 'Powell', 'CG', 'BFGS', 'ANNEAL', 'NELDERMEAD', 'COBYLA', 'BOBYQA', 'PRAXIS', 'SUBPLEX']

for x in methods:
    pyqcm.banner('testing minimization method {:s}'.format(x))
    try:
        run_cdmft(x)
    except:
        print('Method ', x, ' has failed')
